"""Known hosts backend."""
import ipaddress

import pyparsing as pp

from ClusterShell.NodeSet import NodeSet
from ClusterShell.NodeUtils import GroupResolver, GroupSource

from cumin.backends import BaseQueryAggregator, InvalidQueryError


def grammar():
    """Define the query grammar.

    Some query examples:

    * Simple selection: ``host1.domain``
    * ClusterShell syntax for hosts expansion: ``host10[10-42].domain,host2010.other-domain``
    * ClusterShell syntax for hosts globbing: ``host10[10-42]*``
    * A complex selection: ``host100[1-5]* or (host10[30-40].domain and (host10[10-42].domain and not host33.domain))``

    Backus-Naur form (BNF) of the grammar::

        <grammar> ::= <item> | <item> <boolean> <grammar>
           <item> ::= <hosts> | "(" <grammar> ")"
        <boolean> ::= "and not" | "and" | "xor" | "or"

    Given that the pyparsing library defines the grammar in a BNF-like style, for the details of the tokens not
    specified above check directly the source code.

    Returns:
        pyparsing.ParserElement: the grammar parser.

    """
    # Boolean operators
    boolean = (pp.CaselessKeyword('and not').leaveWhitespace() | pp.CaselessKeyword('and')
               | pp.CaselessKeyword('xor') | pp.CaselessKeyword('or'))('bool')

    # Parentheses
    lpar = pp.Literal('(')('open_subgroup')
    rpar = pp.Literal(')')('close_subgroup')

    # Hosts selection: clustershell (,!&^[]) syntax is allowed: host10[10-42].domain
    hosts = (~(boolean) + pp.Word(pp.alphanums + '-_.,!&^[]*?'))('hosts')

    # Final grammar, see the docstring for its BNF based on the tokens defined above
    # Groups are used to split the parsed results for an easy access
    full_grammar = pp.Forward()
    item = hosts | lpar + full_grammar + rpar
    full_grammar << pp.Group(item) + pp.ZeroOrMore(pp.Group(boolean + item))  # pylint: disable=expression-not-assigned

    return full_grammar


class KnownHostsLineError(InvalidQueryError):
    """Custom exception class for invalid lines in SSH known hosts files."""


class KnownHostsSkippedLineError(InvalidQueryError):
    """Custom exception class for skipped lines in SSH known hosts files."""


class KnownHostsQuery(BaseQueryAggregator):
    """KnownHostsQuery query builder.

    The ``knownhosts`` backend allow to use Cumin taking advantage of existing SSH known hosts files that are not
    hashed.
    It allow to write arbitrarily complex queries with subgroups and boolean operators, but each item must be either
    the hostname itself, or using host expansion with the powerful :py:class:`ClusterShell.NodeSet.NodeSet` syntax.

    The typical use case for the ``knownhosts`` backend is when the known hosts file(s) are generated and kept updated
    by some external configuration manager or tool that is not yet supported as a backend for Cumin. It can also work
    as a fallback backend in case the primary backend is unavailable but the known hosts file(s) are still up to date.
    """

    grammar = grammar()
    """:py:class:`pyparsing.ParserElement`: load the grammar parser only once in a singleton-like way."""

    def __init__(self, config):
        """Known hosts query constructor, initialize the known hosts.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery.__init__`.
        """
        super().__init__(config)

        self.known_hosts = set()
        self.resolver = None

    def _build(self, query_string):
        """Override parent method to lazy-loading the known hosts if needed.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._build`.
        """
        if not self.known_hosts:
            self._load_known_hosts()

        if self.resolver is None:
            source = GroupSource('all', allgroups='\n'.join(self.known_hosts))
            self.resolver = GroupResolver(default_source=source)

        super()._build(query_string)

    def _execute(self):
        """Override parent method to ensure to return only existing hosts.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._execute`.
        """
        hosts = super()._execute()
        return hosts & NodeSet('*', resolver=self.resolver)

    def _parse_token(self, token):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQueryAggregator._parse_token`.
        """
        if not isinstance(token, pp.ParseResults):  # pragma: no cover - this should never happen
            raise InvalidQueryError('Expecting ParseResults object, got {type}: {token}'.format(
                type=type(token), token=token))

        token_dict = token.asDict()
        self.logger.trace('Token is: %s | %s', token_dict, token)

        if 'hosts' in token_dict:
            element = self._get_stack_element()
            element['hosts'] = NodeSet.fromlist(token_dict['hosts'], resolver=self.resolver)
            if 'bool' in token_dict:
                element['bool'] = token_dict['bool']
            self.stack_pointer['children'].append(element)
        elif 'open_subgroup' in token_dict and 'close_subgroup' in token_dict:
            self._open_subgroup()
            if 'bool' in token_dict:
                self.stack_pointer['bool'] = token_dict['bool']
            for subtoken in token:
                if isinstance(subtoken, str):  # Grammar literals, boolean operators and parentheses
                    continue
                self._parse_token(subtoken)
            self._close_subgroup()
        else:  # pragma: no cover - this should never happen
            raise InvalidQueryError('Got unexpected token: {token}'.format(token=token))

    def _load_known_hosts(self):
        """Load all known hosts file listed in the configuration."""
        config = self.config.get('knownhosts', {})
        known_hosts_filenames = config.get('files', [])

        for filename in known_hosts_filenames:
            hosts = set()
            with open(filename, 'r') as known_hosts_file:
                for lineno, line in enumerate(known_hosts_file, 1):
                    try:
                        found, skipped = KnownHostsQuery.parse_known_hosts_line(line)
                        if skipped:
                            self.logger.trace("Skipped patterns at line %d in known hosts file '%s': %s",
                                              lineno, filename, ', '.join(skipped))
                        hosts.update(found)
                    except KnownHostsLineError as e:
                        self.logger.warning("Discarded invalid line %d (%s) in known hosts file '%s': %s",
                                            lineno, e, filename, line)
                    except KnownHostsSkippedLineError as e:
                        self.logger.trace("Skipped %s line %d in known hosts file '%s': %s", e, lineno, filename, line)

            self.logger.debug("Loaded %d hosts from '%s'", len(hosts), filename)
            self.known_hosts.update(hosts)

    @staticmethod
    def parse_known_hosts_line(line):
        """Parse an SSH known hosts formatted line and extract the valid hostnames.

        See the ``SSH_KNOWN_HOSTS FILE FORMAT` in ``man sshd`` for the details of the file format.

        Arguments:
            line (str): the line to parse.

        Raises:
            KnownHostsSkippedLineError: if the line is skipped.
            KnownHostsLineError: if unable to parse the line.

        Returns:
            set: a set with the hostnames found in the given line.

        """
        line = line.strip()
        if not line:
            raise KnownHostsSkippedLineError('empty line')

        if line[0] == '#':
            raise KnownHostsSkippedLineError('comment')

        if line[0] == '|':
            raise KnownHostsSkippedLineError('hashed')

        fields = line.split()
        if len(fields) < 3:
            raise KnownHostsLineError('not enough fields')

        if line[0] == '@':
            if len(fields) < 4:
                raise KnownHostsLineError('not enough fields')

            if fields[0] == '@cert-authority':
                line_hosts = fields[1]
            elif fields[0] == '@revoked':
                raise KnownHostsSkippedLineError('revoked')
            else:
                raise KnownHostsLineError('unknown marker')
        else:
            line_hosts = fields[0]

        return KnownHostsQuery.parse_line_hosts(line_hosts)

    @staticmethod
    def parse_line_hosts(line_hosts):
        """Parse a comma-separated hostnamed from an SSH known hosts formatted line and extract the valid hostnames.

        Arguments:
            line_hosts (str): the hostnames to parse.

        Returns:
            tuple: a tuple with two sets, the hostnames found in the given line and the hostnames skipped.

        """
        hosts = set()
        skipped = set()
        for host in line_hosts.split(','):
            if not host:
                continue

            if host[0] == '!':
                host = host[1:]

            if host[0] == '[':
                host = host[1:].split(']')[0]

            if '*' in host or '?' in host:
                skipped.add(host)
            else:
                try:
                    ipaddress.ip_address(host)
                    skipped.add(host)
                except ValueError:
                    hosts.add(host)  # Add hostnames, skip IP addresses

        return hosts, skipped


GRAMMAR_PREFIX = 'K'
""":py:class:`str`: the prefix associate to this grammar, to register this backend into the general grammar.
Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""

query_class = KnownHostsQuery  # pylint: disable=invalid-name
"""Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""



"""Test external backend module with wrong inheritance of the query class."""


class WrongInheritance:
    """Test query class with wrong inheritance."""


GRAMMAR_PREFIX = '_Z'
""":py:class:`str`: the prefix associate to this grammar, to register this backend into the general grammar.
Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""
query_class = WrongInheritance  # pylint: disable=invalid-name
"""Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""



"""OpenStack backend."""
import pyparsing as pp

from keystoneauth1 import session as keystone_session
from keystoneauth1.identity import v3 as keystone_identity
from keystoneclient.v3 import client as keystone_client
from novaclient import client as nova_client

from cumin import nodeset, nodeset_fromlist
from cumin.backends import BaseQuery, InvalidQueryError


def grammar():
    """Define the query grammar.

    Backus-Naur form (BNF) of the grammar::

        <grammar> ::= "*" | <items>
          <items> ::= <item> | <item> <whitespace> <items>
           <item> ::= <key>:<value>

    Given that the pyparsing library defines the grammar in a BNF-like style, for the details of the tokens not
    specified above check directly the source code.

    Returns:
        pyparsing.ParserElement: the grammar parser.

    """
    quoted_string = pp.quotedString.copy().addParseAction(pp.removeQuotes)  # Both single and double quotes are allowed

    # Key-value tokens: key:value
    # Lowercase key, all printable characters except the parentheses that are part of the global grammar for the value
    key = pp.Word(pp.srange('[a-z0-9-_.]"'), min=2)('key')
    all_but_par = ''.join([c for c in pp.printables if c not in ('(', ')', '{', '}')])
    value = (quoted_string | pp.Word(all_but_par))('value')
    item = pp.Combine(key + ':' + value)

    # Final grammar, see the docstring for its BNF based on the tokens defined above
    # Groups are used to split the parsed results for an easy access
    return pp.Group(pp.Literal('*')('all')) | pp.OneOrMore(pp.Group(item))


def _get_keystone_session(config, project=None):
    """Return a new keystone session based on configuration.

    Arguments:
        config (dict): a dictionary with the session configuration keys: ``auth_url``, ``username``, ``password``.
        project (str, optional): a project to scope the session to.

    Returns:
        keystoneauth1.session.Session: the Keystone session scoped for the project if specified.

    """
    auth = keystone_identity.Password(
        auth_url='{auth_url}/v3'.format(auth_url=config.get('auth_url', 'http://localhost:5000')),
        username=config.get('username', 'username'),
        password=config.get('password', 'password'),
        project_name=project,
        user_domain_id='default',
        project_domain_id='default')
    return keystone_session.Session(auth=auth)


def _get_nova_client(config, project):
    """Return a new nova client tailored to the given project.

    Arguments:
        config (dict): a dictionary with the session configuration keys: ``auth_url``, ``username``, ``password``,
            ``nova_api_version``, ``timeout``.
        project (str): the project to scope the `novaclient` session to.

    Returns:
        novaclient.client.Client: the novaclient Client instance, already authenticated.

    """
    params = config.get('client_params', {})
    return nova_client.Client(
        config.get('nova_api_version', '2'),
        session=_get_keystone_session(config, project),
        endpoint_type='public',
        timeout=config.get('timeout', 10),
        **params)


class OpenStackQuery(BaseQuery):
    r"""OpenStackQuery query builder.

    Query VMs deployed in an OpenStack infrastructure using the API.
    This is an optional backend, its dependencies will not be installed automatically, see the Installation section
    of the documentation for more details.

    * Each query can specify multiple parameters to filter the hosts selection in the form ``key:value``.
    * The special ``project`` key allow to filter by the OpenStack project name: ``project:project_name``. If not
      specified all the visible and enabled projects will be queried.
    * Any other ``key:value`` pair will be passed as is to the
      `OpenStack Compute API list-servers <https://developer.openstack.org/api-ref/compute/#list-servers>`_. Multiple
      filters can be added separated by space. The value can be enclosed in single or double quotes:
      ``name:"host1.*\.domain" image:UUID``
    * By default the filters ``status:ACTIVE`` and ``vm_state:ACTIVE`` are also added, but will be overridden if
      specified in the query.
    * To mix multiple selections the general grammar must be used with multiple subqueries:
      ``O{project:project1} or O{project:project2}``
    * The special query ``*`` is a shortcut to select all hosts in all OpenStack projects.
    * See the example configuration in ``doc/examples/config.yaml`` for all the OpenStack-related parameters that can
      be set.

    Some query examples:

    * All hosts in all OpenStack projects: ``*``
    * All hosts in a specific OpenStack project: ``project:project_name``
    * Filter hosts using any parameter allowed by the OpenStack list-servers API: ``name:host1 image:UUID``
      See `OpenStack Compute API list-servers <https://developer.openstack.org/api-ref/compute/#list-servers>`_ for
      more details. Multiple filters can be added separated by space. The value can be enclosed in single or double
      quotes. If the ``project`` key is not specified the hosts will be selected from all projects.
    * To mix multiple selections the general grammar must be used with multiple subqueries:
      ``O{project:project1} or O{project:project2}``
    """

    grammar = grammar()
    """:py:class:`pyparsing.ParserElement`: load the grammar parser only once in a singleton-like way."""

    def __init__(self, config):
        """Override parent class constructor for specific setup.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery.__init__`.

        """
        super().__init__(config)
        self.openstack_config = self.config.get('openstack', {})
        self.search_project = None
        self.search_params = self._get_default_search_params()

    def _get_default_search_params(self):
        """Return the default search parameters dictionary and set the project, if configured.

        Returns:
            dict: the dictionary with the default search parameters.

        """
        params = {'status': 'ACTIVE', 'vm_state': 'ACTIVE'}
        config_params = self.openstack_config.get('query_params', {})

        if 'project' in config_params:
            self.search_project = config_params.pop('project')

        params.update(config_params)
        return params

    def _build(self, query_string):
        """Override parent class _build method to reset the search parameters.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._build`.

        """
        self.search_params = self._get_default_search_params()
        super()._build(query_string)

    def _execute(self):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._execute`.

        Returns:
            ClusterShell.NodeSet.NodeSet: with the FQDNs of the matching hosts.

        """
        if self.search_project is None:
            hosts = nodeset()
            for project in self._get_projects():
                hosts |= self._get_project_hosts(project)
        else:
            hosts = self._get_project_hosts(self.search_project)

        return hosts

    def _parse_token(self, token):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._parse_token`.

        Raises:
            cumin.backends.InvalidQueryError: on internal parsing error.

        """
        if not isinstance(token, pp.ParseResults):  # pragma: no cover - this should never happen
            raise InvalidQueryError('Expecting ParseResults object, got {type}: {token}'.format(
                type=type(token), token=token))

        token_dict = token.asDict()
        self.logger.trace('Token is: %s | %s', token_dict, token)

        if 'key' in token_dict and 'value' in token_dict:
            if token_dict['key'] == 'project':
                self.search_project = token_dict['value']
            else:
                self.search_params[token_dict['key']] = token_dict['value']
        elif 'all' in token_dict:
            pass  # nothing to do, search_project and search_params have the right defaults
        else:  # pragma: no cover - this should never happen
            raise InvalidQueryError('Got unexpected token: {token}'.format(token=token))

    def _get_projects(self):
        """Get all the project names from keystone API, filtering out the special `admin` project. Is a `generator`.

        Yields:
            str: the project name for all the selected projects.

        """
        client = keystone_client.Client(
            session=_get_keystone_session(self.openstack_config), timeout=self.openstack_config.get('timeout', 10))
        return (project.name for project in client.projects.list(enabled=True) if project.name != 'admin')

    def _get_project_hosts(self, project):
        """Return a NodeSet with the list of matching hosts based for the project based on the search parameters.

        Arguments:
            project (str): the project name where to get the list of hosts.

        Returns:
            ClusterShell.NodeSet.NodeSet: with the FQDNs of the matching hosts.

        """
        client = _get_nova_client(self.openstack_config, project)

        domain = ''
        domain_suffix = self.openstack_config.get('domain_suffix', None)
        if domain_suffix is not None:
            if domain_suffix[0] != '.':
                domain = '.{suffix}'.format(suffix=domain_suffix)
            else:
                domain = domain_suffix

        return nodeset_fromlist('{host}.{project}{domain}'.format(host=server.name, project=project, domain=domain)
                                for server in client.servers.list(search_opts=self.search_params))


GRAMMAR_PREFIX = 'O'
""":py:class:`str`: the prefix associate to this grammar, to register this backend into the general grammar.
Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""

query_class = OpenStackQuery  # pylint: disable=invalid-name
"""Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""



"""Grammar tests."""

import os
import sys

from unittest import mock

import pytest

from cumin import CuminError, grammar
from cumin.tests import get_fixture
from cumin.tests.unit.backends.external import ExternalBackendQuery


REGISTERED_BACKENDS = grammar.get_registered_backends()


def test_valid_strings():
    """Run quick pyparsing test over valid grammar strings."""
    results = grammar.grammar(REGISTERED_BACKENDS.keys()).runTests(
        get_fixture(os.path.join('grammar', 'valid_grammars.txt'), as_string=True))
    assert results[0]


def test_invalid_strings():
    """Run quick pyparsing test over invalid grammar strings."""
    results = grammar.grammar(REGISTERED_BACKENDS.keys()).runTests(
        get_fixture(os.path.join('grammar', 'invalid_grammars.txt'), as_string=True), failureTests=True)
    assert results[0]


# Built-in backends registration tests
@mock.patch('cumin.grammar.pkgutil.iter_modules')
def test_duplicate_backend(mocked_iter_modules):
    """Trying to register a backend with the same key of another should raise CuminError."""
    backend = mock.MagicMock()
    backend.GRAMMAR_PREFIX = 'D'
    sys.modules['cumin.backends.test_backend'] = backend
    mocked_iter_modules.return_value = ((None, name, False) for name in ('direct', 'puppetdb', 'test_backend'))
    with pytest.raises(CuminError, match='Unable to register backend'):
        grammar.get_registered_backends()
    del sys.modules['cumin.backends.test_backend']


@mock.patch('cumin.grammar.importlib.import_module')
def test_backend_import_error(mocked_import_modules):
    """If an internal backend raises ImportError because of missing dependencies, it should be skipped."""
    mocked_import_modules.side_effect = ImportError
    backends = grammar.get_registered_backends()
    assert backends == {}


@mock.patch('cumin.grammar.importlib.import_module')
def test_backend_import_error_ext(mocked_import_modules):
    """If an external backend raises ImportError because of missing dependencies, it should raise CuminError."""
    mocked_import_modules.side_effect = ImportError
    with pytest.raises(CuminError, match='Unable to import backend'):
        grammar.get_registered_backends(external=['cumin.tests.unit.backends.external.ok'])


@mock.patch('cumin.grammar.pkgutil.iter_modules')
def test_import_error_backend(mocked_iter_modules):
    """Trying to register a backend that raises ImportError should silently skip it (missing optional dependencies)."""
    # Using a non-existent backend as it will raise ImportError like an existing backend with missing dependencies.
    mocked_iter_modules.return_value = ((None, name, False) for name in ('direct', 'puppetdb', 'non_existent'))
    backends = grammar.get_registered_backends()
    assert len(backends.keys()) == 2
    assert sorted(backends.keys()) == ['D', 'P']


# External backends registration tests
def test_register_ok():
    """An external backend should be properly registered."""
    backends = grammar.get_registered_backends(external=['cumin.tests.unit.backends.external.ok'])
    assert '_Z' in backends
    assert backends['_Z'].keyword == '_Z'
    assert backends['_Z'].name == 'ok'
    assert backends['_Z'].cls == ExternalBackendQuery


def test_register_missing_prefix():
    """Registering an external backend missing the GRAMMAR_PREFIX should raise CuminError."""
    with pytest.raises(CuminError, match='GRAMMAR_PREFIX module attribute not found'):
        grammar.get_registered_backends(external=['cumin.tests.unit.backends.external.missing_grammar_prefix'])


def test_register_duplicate_prefix():
    """Registering an external backend with an already registered GRAMMAR_PREFIX should raise CuminError."""
    with pytest.raises(CuminError, match='already registered'):
        grammar.get_registered_backends(external=['cumin.tests.unit.backends.external.duplicate_prefix'])


def test_register_missing_class():
    """Registering an external backend missing the query_class should raise CuminError."""
    with pytest.raises(CuminError, match='query_class module attribute not found'):
        grammar.get_registered_backends(external=['cumin.tests.unit.backends.external.missing_query_class'])


def test_register_inheritance():
    """Registering an external backend with a query_class with the wrong inheritance should raise CuminError."""
    with pytest.raises(CuminError, match='query_class module attribute is not a subclass'):
        grammar.get_registered_backends(external=['cumin.tests.unit.backends.external.wrong_inheritance'])



"""Test external backend module with missing query class."""


GRAMMAR_PREFIX = '_Z'
""":py:class:`str`: the prefix associate to this grammar, to register this backend into the general grammar.
Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""



"""CLI integration tests."""
# pylint: disable=protected-access


import copy
import json
import os
import re
import sys

import pytest

from cumin import __version__, cli

# Set environment variables
_ENV = {'USER': 'root', 'SUDO_USER': 'user'}
for key, value in _ENV.items():
    os.environ[key] = value

# Dictionary with expected strings to match in the execution stderr:
# {label: string_to_match}
_EXPECTED_LINES = {
    'all_targeted': '5 hosts will be targeted',
    'failed': 'failed',
    'global_timeout': 'global timeout',
    'successfully': 'successfully',
    'dry_run': 'DRY-RUN mode enabled, aborting',
    'subfanout_targeted': '2 hosts will be targeted',
    'ls_success': "100.0% (5/5) success ratio (>= 100.0% threshold) for command: 'ls -la /tmp'.",
    'ls_success_threshold': "100.0% (5/5) success ratio (>= 50.0% threshold) for command: 'ls -la /tmp'.",
    'ls_partial_success': "/5) of nodes failed to execute command 'ls -la /tmp/maybe'",
    'ls_partial_success_ratio_re':
        r"[4-6]0\.0% \([2-3]/5\) success ratio \(< 100\.0% threshold\) for command: 'ls -la /tmp/maybe'\. Aborting.",
    'ls_partial_success_threshold_ratio':
        "60.0% (3/5) success ratio (>= 50.0% threshold) for command: 'ls -la /tmp/maybe'.",
    'ls_failure_batch': "40.0% (2/5) of nodes failed to execute command 'ls -la /tmp/non_existing'",
    'ls_failure_batch_threshold': "80.0% (4/5) of nodes failed to execute command 'ls -la /tmp/non_existing'",
    'ls_total_failure': "100.0% (5/5) of nodes failed to execute command 'ls -la /tmp/non_existing'",
    'ls_total_failure_threshold_ratio':
        "0.0% (0/5) success ratio (< 50.0% threshold) for command: 'ls -la /tmp/non_existing'. Aborting.",
    'date_success': "100.0% (5/5) success ratio (>= 100.0% threshold) for command: 'date'.",
    'date_success_subfanout': "100.0% (2/2) success ratio (>= 100.0% threshold) for command: 'date'.",
    'date_success_threshold': "100.0% (5/5) success ratio (>= 50.0% threshold) for command: 'date'.",
    'date_success_threshold_partial': "60.0% (3/5) success ratio (>= 50.0% threshold) for command: 'date'.",
    'all_success': '100.0% (5/5) success ratio (>= 100.0% threshold) of nodes successfully executed all commands.',
    'all_success_subfanout':
        '100.0% (2/2) success ratio (>= 100.0% threshold) of nodes successfully executed all commands.',
    'all_success_threshold':
        '100.0% (5/5) success ratio (>= 50.0% threshold) of nodes successfully executed all commands.',
    'all_failure':
        '0.0% (0/5) success ratio (< 100.0% threshold) of nodes successfully executed all commands. Aborting.',
    'all_failure_threshold':
        '0.0% (0/5) success ratio (< 50.0% threshold) of nodes successfully executed all commands. Aborting.',
    'global_timeout_executing_re': (r'([2-6]|)0\.0% \([0-3]/5\) of nodes were executing a command when the global '
                                    r'timeout occurred'),
    'global_timeout_executing_threshold_re':
        r'([2-6]|)0\.0% \([0-3]/5\) of nodes were executing a command when the global timeout occurred',
    'global_timeout_pending_re': (r'([2-6]|)0\.0% \([0-3]/5\) of nodes were pending execution when the global timeout '
                                  r'occurred'),
    'global_timeout_pending_threshold_re':
        r'([2-6]|)0\.0% \([0-3]/5\) of nodes were pending execution when the global timeout occurred',
    'sleep_total_failure': "0.0% (0/5) success ratio (< 100.0% threshold) for command: 'sleep 2'. Aborting.",
    'sleep_success': "100.0% (5/5) success ratio (>= 100.0% threshold) for command: 'sleep 0.5'.",
    'sleep_success_threshold': "100.0% (5/5) success ratio (>= 50.0% threshold) for command: 'sleep 0.5'.",
    'sleep_timeout': "100.0% (5/5) of nodes timeout to execute command 'sleep 2'",
    'sleep_timeout_threshold_re': r"[4-8]0\.0% \([2-4]/5\) of nodes timeout to execute command 'sleep 2'",
    'sync': {
        'ls_total_failure_ratio':
            "0.0% (0/5) success ratio (< 100.0% threshold) for command: 'ls -la /tmp/non_existing'. Aborting.",
    },
    'async': {
        'ls_total_failure_ratio':
            "0.0% (0/5) success ratio (< 100.0% threshold). Aborting.",
    },
}

# Tuple of dictionaries with commands to execute for each variant parameters, with the following fields:
# rc: expected return code
# commands: list of commands to execute
# assert_true: list of labels of strings to match with assertTrue() against stderr. [optional]
# assert_false: list of labels of strings to match with assertFalse() against stderr. [optional]
# additional_params: list of additional parameters to pass to the CLI. [optional]
_VARIANTS_COMMANDS = (
    {'rc': 0, 'commands': ['ls -la /tmp', 'date'], 'assert_true': ['all_success'],
     'assert_false': ['failed', 'global_timeout']},
    {'rc': None, 'commands': ['ls -la /tmp/maybe', 'date']},
    {'rc': 2, 'commands': ['ls -la /tmp/non_existing', 'date'], 'assert_true': ['all_failure'],
     'assert_false': ['global_timeout']},
    {'rc': 0, 'commands': ['date', 'date', 'date'], 'assert_true': ['all_success'],
     'assert_false': ['failed', 'global_timeout']},
    {'rc': 2, 'additional_params': ['--global-timeout', '1'], 'commands': ['date', 'sleep 2']},
    {'rc': None, 'additional_params': ['--global-timeout', '1'], 'commands': ['sleep 0.99', 'date']},
    {'rc': 2, 'additional_params': ['-t', '1'], 'commands': ['sleep 2', 'date'],
     'assert_false': ['failed', 'global_timeout', 'date_success']},
    {'rc': 0, 'additional_params': ['-t', '2'], 'commands': ['sleep 0.5', 'date'],
     'assert_false': ['failed', 'global_timeout']},
)

# Tuple of lists of additional parameters to pass to the CLI.
_VARIANTS_PARAMETERS = (
    ['-m', 'sync'],
    ['-m', 'sync', '--batch-size', '2'],
    ['-m', 'sync', '--batch-size', '2', '--batch-sleep', '1.0'],
    ['-m', 'sync', '-p', '50'],
    ['-m', 'sync', '-p', '50', '--batch-size', '2'],
    ['-m', 'sync', '-p', '50', '--batch-size', '2', '--batch-sleep', '1.0'],
    ['-m', 'async'],
    ['-m', 'async', '--batch-size', '2'],
    ['-m', 'async', '--batch-size', '2', '--batch-sleep', '1.0'],
    ['-m', 'async', '-p', '50'],
    ['-m', 'async', '-p', '50', '--batch-size', '2'],
    ['-m', 'async', '-p', '50', '--batch-size', '2', '--batch-sleep', '1.0'],
)

# Expected output for the -o/--out txt option for one node
_TXT_EXPECTED_SINGLE_OUTPUT = """{prefix}{node_id}: First
{prefix}{node_id}: Second
{prefix}{node_id}: Third"""

# Expected output for the -o/--out json option for one node
_JSON_EXPECTED_SINGLE_OUTPUT = 'First\nSecond\nThird'


def make_method(name, commands_set):
    """Method generator with a dynamic name and docstring."""
    params = copy.deepcopy(commands_set)  # Needed to have a different one for each method

    @pytest.mark.variant_params(params)
    def test_variant(self, capsys):
        """Test variant generated function."""
        argv = self.default_params + params['params'] + [self.all_nodes] + params['commands']
        rc = cli.main(argv=argv)
        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)

        if params['rc'] is None:
            params['rc'] = get_rc(params)

        assert rc == params['rc']
        assert _EXPECTED_LINES['all_targeted'] in err, _EXPECTED_LINES['all_targeted']

        labels = params.get('assert_true', [])
        labels += get_global_timeout_expected_lines(params)

        if 'async' in params['params']:
            mode = 'async'
        else:
            mode = 'sync'
            labels += (get_ls_expected_lines(params) + get_date_expected_lines(params)
                       + get_timeout_expected_lines(params))

        for label in labels:
            if label in ('all_success', 'all_failure') and '-p' in params['params']:
                label = '{label}_threshold'.format(label=label)

            if label in _EXPECTED_LINES[mode]:
                string = _EXPECTED_LINES[mode][label]
            else:
                string = _EXPECTED_LINES[label]

            if label.endswith('_re'):
                assert re.search(string, err) is not None, string
            else:
                assert string in err, string

        for label in params.get('assert_false', []):
            assert _EXPECTED_LINES[label] not in err, _EXPECTED_LINES[label]

    # Dynamically set the name and docstring of the generated function to distinguish them
    test_variant.__name__ = 'test_variant_{name}'.format(name=name)
    test_variant.__doc__ = 'variant_function called with params: {params}'.format(params=params)

    return test_variant


def add_variants_methods(indexes):
    """Decorator to add generated tests to a TestClass subclass."""
    def func_wrapper(cls):
        """Dynamic test generator."""
        for i in indexes:
            for j, commands_set in enumerate(_VARIANTS_COMMANDS):
                commands_set['params'] = _VARIANTS_PARAMETERS[i] + commands_set.get('additional_params', [])
                test_input = make_method('params{i:02d}_commands{j:02d}'.format(i=i, j=j), commands_set)
                if test_input.__doc__ is None:
                    raise AssertionError("Missing __doc__ for test {name}".format(name=test_input.__name__))
                setattr(cls, test_input.__name__, test_input)
        return cls

    return func_wrapper


def get_rc(params):
    """Return the expected return code based on the parameters.

    Arguments:
        params: a dictionary with all the parameters passed to the variant_function

    """
    return_value = 2
    if '-p' in params['params'] and '--global-timeout' not in params['params']:
        return_value = 1

    return return_value


def get_global_timeout_expected_lines(params):  # pylint: disable=invalid-name
    """Return a list of expected lines labels for global timeout-based tests.

    Arguments:
        params: a dictionary with all the parameters passed to the variant_function

    """
    expected = []
    if '--global-timeout' not in params['params']:
        return expected

    if '-p' in params['params']:
        expected = ['global_timeout_executing_threshold_re', 'global_timeout_pending_threshold_re']
    else:
        expected = ['global_timeout_executing_re', 'global_timeout_pending_re']

    return expected


def get_timeout_expected_lines(params):
    """Return a list of expected lines labels for timeout-based tests.

    Arguments:
        params: a dictionary with all the parameters passed to the variant_function

    """
    expected = []
    if '-t' not in params['params']:
        return expected

    if params['rc'] == 0:
        # Test successful cases
        if '-p' in params['params']:
            expected = ['sleep_success_threshold', 'date_success_threshold']
        else:
            expected = ['date_success', 'sleep_success']
    else:
        # Test timeout cases
        if '--batch-size' in params['params']:
            expected = ['sleep_timeout_threshold_re']
        else:
            expected = ['sleep_timeout']

    return expected


def get_date_expected_lines(params):
    """Return a list of expected lines labels for the date command based on parameters.

    Arguments:
        params: a dictionary with all the parameters passed to the variant_function

    """
    expected = []
    if 'ls -la /tmp/non_existing' in params['commands']:
        return expected

    if '-p' in params['params']:
        if 'ls -la /tmp/maybe' in params['commands']:
            expected = ['date_success_threshold_partial']
        elif 'ls -la /tmp' in params['commands']:
            expected = ['date_success_threshold']
    elif 'ls -la /tmp' in params['commands']:
        expected = ['date_success']

    return expected


def get_ls_expected_lines(params):
    """Return a list of expected lines labels for the ls command based on the parameters.

    Arguments:
        params: a dictionary with all the parameters passed to the variant_function

    """
    expected = []
    if 'ls -la /tmp' in params['commands']:
        if '-p' in params['params']:
            expected = ['ls_success_threshold']
        else:
            expected = ['ls_success']
    elif 'ls -la /tmp/maybe' in params['commands']:
        if '-p' in params['params']:
            expected = ['ls_partial_success', 'ls_partial_success_threshold_ratio']
        else:
            expected = ['ls_partial_success', 'ls_partial_success_ratio_re']
    elif 'ls -la /tmp/non_existing' in params['commands']:
        if '--batch-size' in params['params']:
            if '-p' in params['params']:
                expected.append('ls_failure_batch_threshold')
            else:
                expected.append('ls_failure_batch')
        else:
            expected.append('ls_total_failure')

        if '-p' in params['params']:
            expected.append('ls_total_failure_threshold_ratio')
        else:
            expected.append('ls_total_failure_ratio')

    return expected


@add_variants_methods(range(len(_VARIANTS_PARAMETERS)))
class TestCLI:
    """CLI module tests."""

    def setup_method(self, _):
        """Set default properties."""
        # pylint: disable=attribute-defined-outside-init
        self.identifier = os.getenv('CUMIN_IDENTIFIER')
        assert self.identifier is not None, 'Unable to find CUMIN_IDENTIFIER environmental variable'
        self.config = os.path.join(os.getenv('CUMIN_TMPDIR', ''), 'config.yaml')
        self.default_params = ['--force', '-d', '-c', self.config]
        self.nodes_prefix = '{identifier}-'.format(identifier=self.identifier)
        self.all_nodes = '{prefix}[1-5]'.format(prefix=self.nodes_prefix)

    def _get_nodes(self, nodes):
        """Return the query for the nodes selection.

        Arguments:
            nodes: a string with the NodeSet nodes selection

        """
        if nodes is None:
            return self.all_nodes

        return '{prefix}[{nodes}]'.format(prefix=self.nodes_prefix, nodes=nodes)

    def test_single_command_subfanout(self, capsys):
        """Executing one command on a subset of nodes smaller than the ClusterShell fanout."""
        params = [self._get_nodes('1-2'), 'date']
        rc = cli.main(argv=self.default_params + params)
        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)
        assert _EXPECTED_LINES['subfanout_targeted'] in err, _EXPECTED_LINES['subfanout_targeted']
        assert _EXPECTED_LINES['date_success_subfanout'] in err, _EXPECTED_LINES['date_success_subfanout']
        assert _EXPECTED_LINES['all_success_subfanout'] in err, _EXPECTED_LINES['all_success_subfanout']
        assert _EXPECTED_LINES['failed'] not in err, _EXPECTED_LINES['failed']
        assert _EXPECTED_LINES['global_timeout'] not in err, _EXPECTED_LINES['global_timeout']
        assert rc == 0

    def test_single_command_supfanout(self, capsys):
        """Executing one command on a subset of nodes greater than the ClusterShell fanout."""
        params = [self.all_nodes, 'date']
        rc = cli.main(argv=self.default_params + params)
        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)
        assert _EXPECTED_LINES['all_targeted'] in err, _EXPECTED_LINES['all_targeted']
        assert _EXPECTED_LINES['date_success'] in err, _EXPECTED_LINES['date_success']
        assert _EXPECTED_LINES['all_success'] in err, _EXPECTED_LINES['all_success']
        assert _EXPECTED_LINES['failed'] not in err, _EXPECTED_LINES['failed']
        assert _EXPECTED_LINES['global_timeout'] not in err, _EXPECTED_LINES['global_timeout']
        assert rc == 0

    def test_dry_run(self, capsys):
        """With --dry-run only the matching hosts are printed."""
        params = ['--dry-run', self.all_nodes, 'date']
        rc = cli.main(argv=self.default_params + params)
        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)
        assert _EXPECTED_LINES['all_targeted'] in err, _EXPECTED_LINES['all_targeted']
        assert _EXPECTED_LINES['dry_run'] in err, _EXPECTED_LINES['dry_run']
        assert _EXPECTED_LINES['successfully'] not in err, _EXPECTED_LINES['successfully']
        assert _EXPECTED_LINES['failed'] not in err, _EXPECTED_LINES['failed']
        assert _EXPECTED_LINES['global_timeout'] not in err, _EXPECTED_LINES['global_timeout']
        assert rc == 0

    def test_timeout(self, capsys):
        """With a timeout shorter than a command it should fail."""
        params = ['--global-timeout', '1', self.all_nodes, 'sleep 2']
        rc = cli.main(argv=self.default_params + params)
        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)
        assert _EXPECTED_LINES['all_targeted'] in err, _EXPECTED_LINES['all_targeted']
        assert re.search(_EXPECTED_LINES['global_timeout_executing_re'], err) is not None, \
            _EXPECTED_LINES['global_timeout_executing_re']
        assert re.search(_EXPECTED_LINES['global_timeout_pending_re'], err) is not None, \
            _EXPECTED_LINES['global_timeout_pending_re']
        assert _EXPECTED_LINES['sleep_total_failure'] in err, _EXPECTED_LINES['sleep_total_failure']
        assert _EXPECTED_LINES['all_failure'] in err, _EXPECTED_LINES['all_failure']
        assert _EXPECTED_LINES['failed'] not in err, _EXPECTED_LINES['failed']
        assert rc == 2

    def test_version(self, capsys):  # pylint: disable=no-self-use
        """Calling --version should return the version and exit."""
        with pytest.raises(SystemExit) as e:
            cli.main(argv=['--version'])

        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)
        assert e.type == SystemExit
        assert e.value.code == 0
        assert err == ''
        assert len(out.splitlines()) == 1
        assert __version__ in out

    def test_out_txt(self, capsys):
        """The -o/--out txt option should print the output expanded for each host, prefixed by the hostname."""
        params = ['-o', 'txt', self.all_nodes, 'cat /tmp/out']
        rc = cli.main(argv=self.default_params + params)
        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)

        assert _EXPECTED_LINES['all_targeted'] in err, _EXPECTED_LINES['all_targeted']
        assert _EXPECTED_LINES['successfully'] in err, _EXPECTED_LINES['successfully']
        assert _EXPECTED_LINES['failed'] not in err, _EXPECTED_LINES['failed']
        assert rc == 0

        expected_out = '\n'.join(
            _TXT_EXPECTED_SINGLE_OUTPUT.format(prefix=self.nodes_prefix, node_id=i) for i in range(1, 6))

        assert out.split(cli.OUTPUT_SEPARATOR + '\n')[1] == expected_out + '\n'

    def test_out_json(self, capsys):
        """The -o/--out json option should print a JSON with hostnames as keys and output as values."""
        params = ['-o', 'json', self.all_nodes, 'cat /tmp/out']
        rc = cli.main(argv=self.default_params + params)
        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)

        assert _EXPECTED_LINES['all_targeted'] in err, _EXPECTED_LINES['all_targeted']
        assert _EXPECTED_LINES['successfully'] in err, _EXPECTED_LINES['successfully']
        assert _EXPECTED_LINES['failed'] not in err, _EXPECTED_LINES['failed']
        assert rc == 0

        expected_out = {self.nodes_prefix + str(i): _JSON_EXPECTED_SINGLE_OUTPUT for i in range(1, 6)}

        assert json.loads(out.split(cli.OUTPUT_SEPARATOR + '\n')[1]) == expected_out



"""CLI tests."""
import argparse

from logging import DEBUG, INFO
from unittest import mock

import pytest

from cumin import cli, CuminError, LOGGING_TRACE_LEVEL_NUMBER, nodeset, transports


# Environment variables
_ENV = {'USER': 'root', 'SUDO_USER': 'user'}
# Command line arguments
_ARGV = ['-c', 'doc/examples/config.yaml', '-d', '-m', 'sync', 'host', 'command1', 'command2']


def _validate_parsed_args(args, no_commands=False):
    """Validate that the parsed args have the proper values."""
    assert args.debug
    assert args.config == 'doc/examples/config.yaml'
    assert args.hosts == 'host'
    if no_commands:
        assert args.dry_run
    else:
        assert args.commands == ['command1', 'command2']


def test_get_parser():
    """Calling get_parser() should return a populated argparse.ArgumentParser object."""
    parser = cli.get_parser()
    assert isinstance(parser, argparse.ArgumentParser)
    assert parser.prog == 'cumin'


def test_parse_args_help(capsys):
    """Calling cumin with -h/--help should return its help message."""
    with pytest.raises(SystemExit) as e:
        cli.parse_args(['-h'])

    out, _ = capsys.readouterr()
    assert e.value.code == 0
    assert 'Cumin CLI - Automation and orchestration framework written in Python' in out


def test_parse_args_ok():
    """A standard set of command line parameters should be properly parsed into their respective variables."""
    args = cli.parse_args(_ARGV)
    _validate_parsed_args(args)


def test_parse_args_no_commands():
    """If no commands are specified, dry-run mode should be implied."""
    args = cli.parse_args(_ARGV[:-2])
    _validate_parsed_args(args, no_commands=True)


def test_parse_args_no_mode():
    """If mode is not specified with multiple commands, parsing the args should raise a parser error."""
    index = _ARGV.index('-m')
    with pytest.raises(SystemExit):
        cli.parse_args(_ARGV[:index] + _ARGV[index + 1:])


def test_target_batch_size():
    """Calling target_batch_size() should properly parse integer values."""
    assert cli.target_batch_size('1') == {'value': 1, 'ratio': None}
    assert cli.target_batch_size('100') == {'value': 100, 'ratio': None}


@pytest.mark.parametrize('percentage', (('0%', 0.0), ('50%', 0.5), ('100%', 1.0)))
def test_target_batch_size_perc(percentage):
    """Calling target_batch_size() with a valid percentage should properly parse it."""
    assert cli.target_batch_size(percentage[0]) == {'value': None, 'ratio': percentage[1]}


@pytest.mark.parametrize('percentage', ('-1%', '101%'))
def test_target_batch_size_perc_ko(percentage):
    """Calling target_batch_size() with invalid percentage should raise argparse.ArgumentTypeError."""
    with pytest.raises(argparse.ArgumentTypeError, match='not a valid percentage'):
        cli.target_batch_size(percentage)


@pytest.mark.parametrize('value', ('0', '-1'))
def test_target_batch_size_val_ko(value):
    """Calling target_batch_size() with invalid value should raise argparse.ArgumentTypeError."""
    with pytest.raises(argparse.ArgumentTypeError, match='is not a valid value'):
        cli.target_batch_size(value)


def test_get_running_user():
    """Unknown user should raise CuminError and a proper user should be detected."""
    env = {'USER': 'root', 'SUDO_USER': None}
    with mock.patch('os.getenv', env.get):
        with pytest.raises(CuminError, match='Unable to determine real user'):
            cli.get_running_user()

    env = {'USER': 'user', 'SUDO_USER': None}
    with mock.patch('os.getenv', env.get):
        assert cli.get_running_user() == 'user'

    with mock.patch('os.getenv', _ENV.get):
        assert cli.get_running_user() == 'user'


@mock.patch('cumin.cli.os.path.exists')
@mock.patch('cumin.cli.os.makedirs')
@mock.patch('cumin.cli.RotatingFileHandler')
@mock.patch('cumin.cli.logging.getLogger')
def test_setup_logging(mocked_get_logger, mocked_file_handler, mocked_os_makedirs, mocked_os_path_exists):
    """Calling setup_logging() should properly setup the logger."""
    mocked_os_path_exists.return_value = False
    cli.setup_logging('/path/to/filename.yaml')
    assert mock.call().setLevel(INFO) in mocked_get_logger.mock_calls
    assert mocked_file_handler.called
    assert mocked_os_makedirs.called
    assert mocked_os_path_exists.called

    mocked_file_handler.reset_mock()
    mocked_os_makedirs.reset_mock()
    mocked_os_path_exists.reset_mock()

    mocked_os_path_exists.side_effect = FileNotFoundError
    cli.setup_logging('filename.yaml')
    assert mock.call().setLevel(INFO) in mocked_get_logger.mock_calls
    assert mocked_file_handler.called
    assert not mocked_os_makedirs.called
    assert not mocked_os_path_exists.called

    mocked_os_path_exists.return_value = True
    cli.setup_logging('filename.yaml', debug=True)
    assert mock.call().setLevel(DEBUG) in mocked_get_logger.mock_calls

    mocked_os_path_exists.return_value = True
    cli.setup_logging('filename.yaml', trace=True)
    assert mock.call().setLevel(LOGGING_TRACE_LEVEL_NUMBER) in mocked_get_logger.mock_calls


@mock.patch('cumin.cli.stderr')
@mock.patch('builtins.input')
@mock.patch('cumin.cli.sys.stdout.isatty')
@mock.patch('cumin.cli.logger')
def test_sigint_handler(logging, isatty, mocked_input, stderr):  # pylint: disable=unused-argument
    """Calling the SIGINT handler should raise KeyboardInterrupt or not based on tty and answer."""
    # Signal handler called without a tty
    isatty.return_value = False
    with pytest.raises(cli.KeyboardInterruptError):
        cli.sigint_handler(1, None)

    # Signal handler called with a tty
    isatty.return_value = True
    with pytest.raises(cli.KeyboardInterruptError):
        cli.sigint_handler(1, None)

    # # Signal handler called with a tty, answered 'y'
    # isatty.return_value = True
    # mocked_input.return_value = 'y'
    # with pytest.raises(cli.KeyboardInterruptError):
    #     cli.sigint_handler(1, None)
    #
    # # Signal handler called with a tty, answered 'n'
    # isatty.return_value = True
    # mocked_input.return_value = 'n'
    # assert cli.sigint_handler(1, None) is None
    #
    # # Signal handler called with a tty, answered 'invalid_answer'
    # isatty.return_value = True
    # mocked_input.return_value = 'invalid_answer'
    # with pytest.raises(cli.KeyboardInterruptError):
    #     cli.sigint_handler(1, None)
    #
    # # Signal handler called with a tty, empty answer
    # isatty.return_value = True
    # mocked_input.return_value = ''
    # with pytest.raises(cli.KeyboardInterruptError):
    #     cli.sigint_handler(1, None)


@mock.patch('cumin.cli.tqdm')
def test_stderr(tqdm):
    """Calling stderr() should call tqdm.write()."""
    cli.stderr('message')
    assert tqdm.write.called


@mock.patch('cumin.cli.stderr')
@mock.patch('builtins.input')
@mock.patch('cumin.cli.sys.stdout.isatty')
def test_get_hosts_ok(isatty, mocked_input, stderr):
    """Calling get_hosts() should query the backend and return the list of hosts."""
    args = cli.parse_args(['D{host1}', 'command1'])
    config = {'backend': 'direct'}
    isatty.return_value = True

    mocked_input.return_value = 'y'
    assert cli.get_hosts(args, config) == nodeset('host1')

    mocked_input.return_value = 'n'
    with pytest.raises(cli.KeyboardInterruptError):
        cli.get_hosts(args, config)

    mocked_input.return_value = 'invalid_answer'
    with pytest.raises(cli.KeyboardInterruptError):
        cli.get_hosts(args, config)

    mocked_input.return_value = ''
    with pytest.raises(cli.KeyboardInterruptError):
        cli.get_hosts(args, config)

    assert stderr.called


@mock.patch('cumin.cli.stderr')
@mock.patch('cumin.cli.sys.stdout.isatty')
def test_get_hosts_no_tty_ko(isatty, stderr):
    """Calling get_hosts() without a TTY should raise CuminError if --dry-run or --force are not specified."""
    args = cli.parse_args(['D{host1}', 'command1'])
    config = {'backend': 'direct'}
    isatty.return_value = False
    with pytest.raises(CuminError, match='Not in a TTY but neither DRY-RUN nor FORCE mode were specified'):
        cli.get_hosts(args, config)
    assert stderr.called


@mock.patch('cumin.cli.stderr')
@mock.patch('cumin.cli.sys.stdout.isatty')
def test_get_hosts_no_tty_dry_run(isatty, stderr):
    """Calling get_hosts() with or without a TTY with --dry-run should return an empty list."""
    args = cli.parse_args(['--dry-run', 'D{host1}', 'command1'])
    config = {'backend': 'direct'}
    assert cli.get_hosts(args, config) == []
    isatty.return_value = True
    assert cli.get_hosts(args, config) == []
    assert stderr.called


@mock.patch('cumin.cli.stderr')
@mock.patch('cumin.cli.sys.stdout.isatty')
def test_get_hosts_no_tty_force(isatty, stderr):
    """Calling get_hosts() with or without a TTY with --force should return the list of hosts."""
    args = cli.parse_args(['--force', 'D{host1}', 'command1'])
    config = {'backend': 'direct'}
    assert cli.get_hosts(args, config) == nodeset('host1')
    isatty.return_value = True
    assert cli.get_hosts(args, config) == nodeset('host1')
    assert stderr.called


@mock.patch('cumin.cli.cumin.transport.Transport')
@mock.patch('cumin.cli.stderr')
def test_run(stderr, transport):
    """Calling run() should query the hosts and execute the commands on the transport."""
    args = cli.parse_args(['--force', 'D{host1}', 'command1'])
    config = {'backend': 'direct', 'transport': 'clustershell'}
    cli.run(args, config)
    assert transport.new.call_args[0][0] is config
    assert isinstance(transport.new.call_args[0][1], transports.Target)
    assert stderr.called


def test_validate_config_valid():
    """A valid config should be validated without raising exception."""
    cli.validate_config({'log_file': '/var/log/cumin/cumin.log'})


def test_validate_config_invalid():
    """An invalid config should raise CuminError."""
    with pytest.raises(CuminError, match='Missing required parameter'):
        cli.validate_config({})



"""Transport class tests."""

import pkgutil

from unittest import mock

import pytest

from cumin import CuminError, transports
from cumin.transport import Transport


def test_missing_transport():
    """Not passing a transport should raise CuminError."""
    with pytest.raises(CuminError, match=r"Missing required parameter 'transport'"):
        Transport.new({}, transports.Target(['host1']))


def test_invalid_transport():
    """Passing an invalid transport should raise CuminError."""
    with pytest.raises(CuminError, match=r"No module named 'cumin\.transports\.non_existent_transport'"):
        Transport.new({'transport': 'non_existent_transport'}, transports.Target(['host1']))


def test_missing_worker_class():
    """Passing a transport without a defined worker_class should raise CuminError."""
    module = mock.MagicMock()
    del module.worker_class
    with mock.patch('importlib.import_module', lambda _: module):
        with pytest.raises(CuminError, match=r'worker_class'):
            Transport.new({'transport': 'invalid_transport'}, transports.Target(['host1']))


@pytest.mark.parametrize('transport', [name for _, name, ispkg in pkgutil.iter_modules(transports.__path__)
                                       if not ispkg])
def test_valid_transport(transport):
    """Passing a valid transport should return an instance of BaseWorker."""
    assert isinstance(Transport.new({'transport': transport}, transports.Target(['host1'])), transports.BaseWorker)



"""OpenStack backend tests."""
from collections import namedtuple
from unittest import mock

from cumin import nodeset
from cumin.backends import BaseQuery, openstack


Project = namedtuple('Project', ['name'])
Server = namedtuple('Server', ['name'])


def test_openstack_query_class():
    """An instance of query_class should be an instance of BaseQuery."""
    query = openstack.query_class({})
    assert isinstance(query, BaseQuery)


def test_openstack_query_class_init():
    """An instance of OpenStackQuery should be an instance of BaseQuery."""
    config = {'key': 'value'}
    query = openstack.OpenStackQuery(config)
    assert isinstance(query, BaseQuery)
    assert query.config == config


def test_all_selection():
    """A selection for all hosts is properly parsed and interpreted."""
    parsed = openstack.grammar().parseString('*', parseAll=True)
    assert parsed[0].asDict() == {'all': '*'}


def test_key_value_token():
    """A token is properly parsed and interpreted."""
    parsed = openstack.grammar().parseString('project:project_name', parseAll=True)
    assert parsed[0].asDict() == {'key': 'project', 'value': 'project_name'}


def test_key_value_tokens():
    """Multiple tokens are properly parsed and interpreted."""
    parsed = openstack.grammar().parseString('project:project_name name:hostname', parseAll=True)
    assert parsed[0].asDict() == {'key': 'project', 'value': 'project_name'}
    assert parsed[1].asDict() == {'key': 'name', 'value': 'hostname'}


@mock.patch('cumin.backends.openstack.nova_client.Client')
@mock.patch('cumin.backends.openstack.keystone_client.Client')
@mock.patch('cumin.backends.openstack.keystone_session.Session')
@mock.patch('cumin.backends.openstack.keystone_identity.Password')
class TestOpenStackQuery:
    """OpenStack backend query test class."""

    def setup_method(self, _):
        """Set an instance of OpenStackQuery for each test."""
        self.config = {'openstack': {}}  # pylint: disable=attribute-defined-outside-init
        self.query = openstack.OpenStackQuery(self.config)  # pylint: disable=attribute-defined-outside-init

    def test_execute_all(self, keystone_identity, keystone_session, keystone_client, nova_client):
        """Calling execute() with a query that select all hosts should return the list of all hosts."""
        keystone_client.return_value.projects.list.return_value = [Project('project1'), Project('project2')]
        nova_client.return_value.servers.list.side_effect = [
            [Server('host1'), Server('host2')], [Server('host1'), Server('host2')]]

        hosts = self.query.execute('*')
        assert hosts == nodeset('host[1-2].project[1-2]')

        assert keystone_identity.call_count == 3
        assert keystone_session.call_count == 3
        keystone_client.assert_called_once_with(session=keystone_session(), timeout=10)
        assert nova_client.call_args_list == [
            mock.call('2', endpoint_type='public', session=keystone_session(), timeout=10),
            mock.call('2', endpoint_type='public', session=keystone_session(), timeout=10)]
        assert nova_client().servers.list.call_args_list == [
            mock.call(search_opts={'vm_state': 'ACTIVE', 'status': 'ACTIVE'})] * 2

    def test_execute_project(self, keystone_identity, keystone_session, keystone_client, nova_client):
        """Calling execute() with a query that select all hosts in a project should return the list of hosts."""
        nova_client.return_value.servers.list.return_value = [Server('host1'), Server('host2')]

        hosts = self.query.execute('project:project1')
        assert hosts == nodeset('host[1-2].project1')

        assert keystone_identity.call_count == 1
        assert keystone_session.call_count == 1
        keystone_client.assert_not_called()
        nova_client.assert_called_once_with('2', endpoint_type='public', session=keystone_session(), timeout=10)
        nova_client().servers.list.assert_called_once_with(search_opts={'vm_state': 'ACTIVE', 'status': 'ACTIVE'})

    def test_execute_project_name(self, keystone_identity, keystone_session, keystone_client, nova_client):
        """Calling execute() with a query that select hosts matching a name in a project should return only those."""
        nova_client.return_value.servers.list.return_value = [Server('host1'), Server('host2')]

        hosts = self.query.execute('project:project1 name:host')
        assert hosts == nodeset('host[1-2].project1')

        assert keystone_identity.call_count == 1
        assert keystone_session.call_count == 1
        keystone_client.assert_not_called()
        nova_client.assert_called_once_with('2', endpoint_type='public', session=keystone_session(), timeout=10)
        nova_client().servers.list.assert_called_once_with(
            search_opts={'vm_state': 'ACTIVE', 'status': 'ACTIVE', 'name': 'host'})

    def test_execute_project_domain(self, keystone_identity, keystone_session, keystone_client, nova_client):
        """When the domain suffix is configured, it should append it to all hosts."""
        nova_client.return_value.servers.list.return_value = [Server('host1'), Server('host2')]
        self.config['openstack']['domain_suffix'] = 'servers.local'
        query = openstack.OpenStackQuery(self.config)

        hosts = query.execute('project:project1')
        assert hosts == nodeset('host[1-2].project1.servers.local')

        assert keystone_identity.call_count == 1
        assert keystone_session.call_count == 1
        keystone_client.assert_not_called()

    def test_execute_project_dot_domain(self, keystone_identity, keystone_session, keystone_client, nova_client):
        """When the domain suffix is configured with a dot, it should append it to all hosts without the dot."""
        nova_client.return_value.servers.list.return_value = [Server('host1'), Server('host2')]
        self.config['openstack']['domain_suffix'] = '.servers.local'
        query = openstack.OpenStackQuery(self.config)

        hosts = query.execute('project:project1')
        assert hosts == nodeset('host[1-2].project1.servers.local')

        assert keystone_identity.call_count == 1
        assert keystone_session.call_count == 1
        keystone_client.assert_not_called()

    def test_execute_query_params(self, keystone_identity, keystone_session, keystone_client, nova_client):
        """When the query_params are set, they must be loaded automatically."""
        nova_client.return_value.servers.list.return_value = [Server('host1'), Server('host2')]
        self.config['openstack']['query_params'] = {'project': 'project1'}
        query = openstack.OpenStackQuery(self.config)

        hosts = query.execute('*')
        assert hosts == nodeset('host[1-2].project1')

        assert keystone_identity.call_count == 1
        assert keystone_session.call_count == 1
        keystone_client.assert_not_called()



"""Pytest customization for integration tests."""
import pytest


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):  # pylint: disable=unused-argument
    """If a custom variant_params marker is set, print a section with its content."""
    outcome = yield
    marker = item.get_closest_marker('variant_params')
    if marker:
        rep = outcome.get_result()
        rep.sections.insert(0, ('test_variant parameters', marker.args))



#!/usr/bin/python3
"""Cumin CLI entry point."""
import argparse
import code
import json
import logging
import os
import pkgutil
import signal
import sys

from logging.handlers import RotatingFileHandler  # pylint: disable=ungrouped-imports

from tqdm import tqdm

import cumin

from cumin import backends, query, transport, transports
from cumin.color import Colored


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
"""logging.Logger: The logging instance."""
OUTPUT_FORMATS = ('txt', 'json')
"""tuple: A tuple with the possible output formats."""
OUTPUT_SEPARATOR = '_____FORMATTED_OUTPUT_____'
"""str: The output separator used when -o/--out is set."""
INTERACTIVE_BANNER = """===== Cumin Interactive REPL =====
# Press Ctrl+d or type exit() to exit the program.

= Available variables =
#     hosts: the ClusterShell NodeSet of targeted hosts.
#     worker: the instance of the Transport worker that was used for the execution.
#     args: the parsed command line arguments, an argparse.Namespace instance.
#     config: the cofiguration dictionary.
#     exit_code: the return code of the execution, that will be used as exit code.

= Useful functions =
#     worker.get_results(): generator that yields the tuple (nodes, output) for each grouped result, where:
#                         -     nodes: is a ClusterShell.NodeSet.NodeSet instance
#                         -     output: is a ClusterShell.MsgTree.MsgTreeElem instance
#     h(): print this help message.
#     help(object_name): Python default interactive help and documentation of the given object.

= Example usage:
for nodes, output in worker.get_results():
    print(nodes)
    print(output.message().decode())
    print('-----')
"""
"""str: The message to print when entering the intractive REPL mode."""


class KeyboardInterruptError(cumin.CuminError):
    """Custom KeyboardInterrupt exception class for the SIGINT signal handler."""


def get_parser():
    """Create and return the command line arguments parser.

    Returns:
        argparse.ArgumentParser: the parser object.

    """
    sync_mode = 'sync'
    async_mode = 'async'

    # Get the list of existing backends and transports
    backends_names = [name for _, name, ispkg in pkgutil.iter_modules(backends.__path__) if not ispkg]
    transports_names = [name for _, name, ispkg in pkgutil.iter_modules(transports.__path__) if not ispkg]

    parser = argparse.ArgumentParser(
        prog='cumin',
        description='Cumin CLI - Automation and orchestration framework written in Python',
        epilog='More details at https://wikitech.wikimedia.org/wiki/Cumin')
    parser.add_argument('-c', '--config', default='/etc/cumin/config.yaml',
                        help='configuration file. [default: /etc/cumin/config.yaml]')
    parser.add_argument('--global-timeout', type=int,
                        help='Global timeout in seconds (int) for the whole execution. [default: None (unlimited)]')
    parser.add_argument('-t', '--timeout', type=int,
                        help=('Timeout in seconds (int) for the the execution of every command in each host. '
                              '[default: None (unlimited)]'))
    parser.add_argument('-m', '--mode', choices=(sync_mode, async_mode),
                        help=('Execution mode, required when there are multiple COMMANDS to be executed. In sync mode, '
                              'execute the first command on all hosts, then proceed with the next one only if '
                              '-p/--success-percentage is reached. In async mode, execute on each host independently '
                              'from each other, the list of commands, aborting the execution on any given host at the '
                              'first command that fails.'))
    parser.add_argument('-p', '--success-percentage', type=int, choices=range(101), metavar='PCT', default=100,
                        help=(('Percentage threshold to consider an execution unit successful. Required in sync mode, '
                               'optional in async mode when -b/--batch-size is used. Accepted values are integers '
                               'in the range 0-100. [default: 100]')))
    parser.add_argument('-b', '--batch-size', type=target_batch_size, default={'value': None, 'ratio': None},
                        help=('The commands will be executed with a sliding batch of this size. The batch mode depends '
                              'on the -m/--mode option when multiple commands are specified. In sync mode the first '
                              'command is executed in batch to all hosts before proceeding with the next one. In async '
                              'mode all commands are executed on the first batch of hosts, proceeding with the next '
                              'hosts as soon as one host completes all the commands. The -p/--success-percentage is '
                              'checked before starting the execution in each host. It accept an absolute integer '
                              '(i.e. 10) or a percentage (i.e. 50%%). [default: None (# of hosts)]'))
    parser.add_argument('-s', '--batch-sleep', type=float,
                        help=('Sleep in seconds (float) to wait before starting the execution on the next host when '
                              '-b/--batch-size is used. [default: None]'))
    parser.add_argument('-x', '--ignore-exit-codes', action='store_true',
                        help=('USE WITH CAUTION! Treat any executed command as successful, ignoring the exit codes. '
                              '[default: False]'))
    parser.add_argument('-o', '--output', choices=OUTPUT_FORMATS,
                        help='Specify a different output format. [default: None]')
    parser.add_argument('-i', '--interactive', action='store_true',
                        help='Drop into a Python shell with the results. [default: False]')
    parser.add_argument('--force', action='store_true',
                        help=('USE WITH CAUTION! Force the execution without confirmation of the affected hosts. '
                              '[default: False]'))
    parser.add_argument('--backend',
                        help=('Override the default backend selected in the configuration file for this execution. The '
                              'backend-specific configuration must be already present in the configuration file. '
                              'One of [{backends}] or any external backend listed in the configuration file '
                              '[default: None]').format(backends=', '.join(backends_names)))
    parser.add_argument('--transport', choices=transports_names,
                        help=('Override the default transport selected in the configuration file for this execution. '
                              'The transport-specific configuration must already be present in the configuration file. '
                              '[default: None]'))
    parser.add_argument('--dry-run', action='store_true',
                        help=('Do not execute any command, just return the list of matching hosts and exit. '
                              '[default: False]'))
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=cumin.__version__))
    parser.add_argument('-d', '--debug', action='store_true',
                        help=('Set log level to DEBUG. See also log_file in the configuration. [default: False]'))
    parser.add_argument('--trace', action='store_true',
                        help=('Set log level to TRACE, a custom logging level intended for development debugging. See '
                              'also log_file in the configuration. [default: False]'))
    parser.add_argument('hosts', metavar='HOSTS_QUERY', help='Hosts selection query')
    parser.add_argument('commands', metavar='COMMAND', nargs='*',
                        help='Command to be executed. If no commands are specified, --dry-run is set.')

    return parser


def target_batch_size(string):
    """Validator for the --batch-size command line argument to be used as type in ArgumentParser.

    Arguments:
        string: the input string to be validated and parsed.

    Returns:
        dict: a dictionary with the batch size absolute value as integer (`value` key) or the ratio value as float
            (`ratio` key) to be used when instantiating a Target object.

    """
    is_percentage = False
    orig = string
    if string[-1] == '%':
        is_percentage = True
        string = string[:-1]

    value = int(string)
    if is_percentage and not (0 <= value <= 100):  # pylint: disable=superfluous-parens
        raise argparse.ArgumentTypeError(
            '{size} is not a valid percentage, expected in range 0%-100% or positive integer.'.format(size=orig))

    if not is_percentage and value <= 0:
        raise argparse.ArgumentTypeError(
            '{size} is not a valid value, expected positive integer or percentage in range 0%-100%'.format(size=orig))

    ret = {'value': None, 'ratio': None}
    if is_percentage:
        ret['ratio'] = value / 100
    else:
        ret['value'] = value

    return ret


def parse_args(argv):
    """Parse command line arguments, validate and return them.

    Arguments:
        argv: the list of command line arguments to use.

    Returns:
        argparse.Namespace: the parsed arguments.

    """
    parser = get_parser()
    parsed_args = parser.parse_args(argv)

    # Validation and default values
    num_commands = len(parsed_args.commands)
    if num_commands == 0:
        parsed_args.dry_run = True
    elif num_commands == 1:
        parsed_args.mode = 'sync'
    elif num_commands > 1:
        if parsed_args.mode is None:
            parser.error('-m/--mode is required when there are multiple COMMANDS')
        if parsed_args.interactive:
            parser.error('-i/--interactive can be used only with one command')
        if parsed_args.output is not None:
            parser.error('-o/--output can be used only with one command')

    if parsed_args.ignore_exit_codes:
        stderr('IGNORE EXIT CODES mode enabled, all commands executed will be considered successful')

    return parsed_args


def get_running_user():
    """Ensure that the original user is detected and return it."""
    if os.getenv('USER') == 'root':
        if os.getenv('SUDO_USER') in (None, 'root'):
            raise cumin.CuminError('Unable to determine real user, logged in as root?')
        return os.getenv('SUDO_USER')
    return os.getenv('USER')


def setup_logging(filename, debug=False, trace=False):
    """Setup the logger instance.

    Arguments:
        filename: the filename of the log file
        debug: whether to set logging level to DEBUG [optional, default: False]
        trace: whether to set logging level to TRACE [optional, default: False]

    """
    file_path = os.path.dirname(filename)
    if file_path and not os.path.exists(file_path):
        os.makedirs(file_path, 0o770)

    log_formatter = logging.Formatter(fmt='%(asctime)s [%(levelname)s %(process)s %(name)s.%(funcName)s] %(message)s')
    log_handler = RotatingFileHandler(filename, maxBytes=(5 * (1024**2)), backupCount=30)
    log_handler.setFormatter(log_formatter)

    root_logger = logging.getLogger()
    root_logger.addHandler(log_handler)
    root_logger.raiseExceptions = False

    if trace:
        root_logger.setLevel(cumin.LOGGING_TRACE_LEVEL_NUMBER)
    elif debug:
        root_logger.setLevel(logging.DEBUG)
    else:
        root_logger.setLevel(logging.INFO)


def sigint_handler(*args):  # pylint: disable=unused-argument
    """Signal handler for Ctrl+c / SIGINT, raises KeyboardInterruptError.

    Arguments (as defined in https://docs.python.org/3/library/signal.html):
        signum: the signal number
        frame: the current stack frame
    """
    if not sys.stdout.isatty():  # pylint: disable=no-member
        logger.warning('Execution interrupted by Ctrl+c/SIGINT')
        raise KeyboardInterruptError

    # TODO: make the below code block to work as expected with ClusterShell
    #       temporarily exit upon Ctrl+c also in interactive mode
    logger.warning('Execution interrupted by Ctrl+c/SIGINT')
    raise KeyboardInterruptError
    # logger.warning('Received Ctrl+c/SIGINT')
    # for i in xrange(10):
    #     stderr('Ctrl+c pressed, sure to quit [y/n]?\n')
    #     try:
    #         answer = input('\n')  # nosec
    #     except RuntimeError:
    #         # Can't re-enter readline when already waiting for input in get_hosts(). Assuming 'y' as answer
    #         stderr('Ctrl+c pressed while waiting for answer. Aborting')
    #         answer = 'y'
    #
    #     if not answer:
    #         continue
    #
    #     if answer.lower() == 'y':
    #         logger.warning('Execution interrupted by Ctrl+c/SIGINT')
    #         raise KeyboardInterruptError
    #     elif answer.lower() == 'n':
    #         message = 'Ctrl+c/SIGINT aborted, resuming execution'
    #         logger.warning(message)
    #         stderr(message)
    #         break
    # else:
    #     logger.warning('Execution interrupted by Ctrl+c/SIGINT: got invalid answer for %d times', i)
    #     raise KeyboardInterruptError


def stderr(message, end='\n'):
    r"""Print a message to stderr and flush.

    Arguments:
        message: the message to print to sys.stderr
        end: the character to use at the end of the message. [optional, default: \n]

    """
    tqdm.write(Colored.yellow(message), file=sys.stderr, end=end)


def get_hosts(args, config):
    """Resolve the hosts selection into a list of hosts and return it. Raises KeyboardInterruptError.

    Arguments:
        args: ArgumentParser instance with parsed command line arguments
        config: a dictionary with the parsed configuration file

    """
    hosts = query.Query(config).execute(args.hosts)

    if not hosts:
        stderr('No hosts found that matches the query')
        return hosts

    stderr('{num} hosts will be targeted:'.format(num=len(hosts)))
    stderr(Colored.cyan(cumin.nodeset_fromlist(hosts)))

    if args.dry_run:
        stderr('DRY-RUN mode enabled, aborting')
        return []

    if args.force:
        stderr('FORCE mode enabled, continuing without confirmation')
        return hosts

    if not sys.stdout.isatty():  # pylint: disable=no-member
        message = 'Not in a TTY but neither DRY-RUN nor FORCE mode were specified.'
        stderr(message)
        raise cumin.CuminError(message)

    for i in range(10):
        stderr('Confirm to continue [y/n]?', end=' ')
        answer = input()  # nosec
        if not answer:
            continue

        if answer in 'yY':
            break
        if answer in 'nN':
            raise KeyboardInterruptError

    else:
        stderr('Got invalid answer for {i} times'.format(i=i))
        raise KeyboardInterruptError

    return hosts


def print_output(output_format, worker):
    """Print the execution results in a specific format.

    Arguments:
        output_format: the output format to use, one of: 'txt', 'json'.
        worker: the Transport worker instance to retrieve the results from.

    """
    if output_format not in OUTPUT_FORMATS:
        raise cumin.CuminError("Got invalid output format '{fmt}', expected one of {allowed}".format(
            fmt=output_format, allowed=OUTPUT_FORMATS))

    out = {}
    for nodeset, output in worker.get_results():
        for node in nodeset:
            if output_format == 'txt':
                out[node] = '\n'.join('{node}: {line}'.format(node=node, line=line.decode()) for line in output.lines())
            elif output_format == 'json':
                out[node] = output.message().decode()

    if output_format == 'txt':
        for node in sorted(out.keys()):
            tqdm.write(out[node])
    elif output_format == 'json':
        tqdm.write(json.dumps(out, indent=4, sort_keys=True))


def run(args, config):
    """Execute the commands on the selected hosts and print the results.

    Arguments:
        args: ArgumentParser instance with parsed command line arguments
        config: a dictionary with the parsed configuration file

    """
    hosts = get_hosts(args, config)
    if not hosts:
        return 0

    target = transports.Target(hosts, batch_size=args.batch_size['value'], batch_size_ratio=args.batch_size['ratio'],
                               batch_sleep=args.batch_sleep)
    worker = transport.Transport.new(config, target)

    ok_codes = None
    if args.ignore_exit_codes:
        ok_codes = []

    worker.commands = [transports.Command(command, timeout=args.timeout, ok_codes=ok_codes)
                       for command in args.commands]
    worker.timeout = args.global_timeout
    worker.handler = args.mode
    worker.success_threshold = args.success_percentage / 100
    exit_code = worker.execute()

    if args.interactive:
        # Define a help function h() that will be available in the interactive shell to print the help message.
        # The name is to not shadow the Python built-in help() that might be usefult too to inspect objects.
        def h():  # pylint: disable=possibly-unused-variable,invalid-name
            """Print the help message in interactive shell."""
            tqdm.write(INTERACTIVE_BANNER)
        code.interact(banner=INTERACTIVE_BANNER, local=locals())
    elif args.output is not None:
        tqdm.write(OUTPUT_SEPARATOR)
        print_output(args.output, worker)

    return exit_code


def validate_config(config):
    """Perform validation of mandatory keys in the configuration.

    Arguments:
        config (dict): the loaded configuration to validate.

    Raises:
        cumin.CuminError: if a mandatory configuration key is missing.

    """
    if 'log_file' not in config:
        raise cumin.CuminError("Missing required parameter 'log_file' in the configuration file '{config}'".format(
            config=config))


def main(argv=None):
    """CLI entry point. Execute commands on hosts according to arguments.

    Arguments:
        argv: the list of command line arguments to use. If not specified it will be automatically taken from sys.argv
            [optional, default: None]

    """
    if argv is None:
        argv = sys.argv[1:]

    signal.signal(signal.SIGINT, sigint_handler)

    # Setup
    try:
        args = parse_args(argv)
        user = get_running_user()
        config = cumin.Config(args.config)
        validate_config(config)
        setup_logging(config['log_file'], debug=args.debug, trace=args.trace)
    except cumin.CuminError as e:
        stderr(e)
        return 2
    except Exception as e:  # pylint: disable=broad-except
        stderr('Caught {name} exception: {msg}'.format(name=e.__class__.__name__, msg=e))
        return 3

    # Override config with command line arguments
    if args.backend is not None:
        config['default_backend'] = args.backend
    if args.transport is not None:
        config['transport'] = args.transport

    logger.info("Cumin called by user '%s' with args: %s", user, args)

    # Execution
    try:
        exit_code = run(args, config)
    except KeyboardInterruptError:
        stderr('Execution interrupted by Ctrl+c/SIGINT/Aborted')
        exit_code = 98
    except Exception as e:  # pylint: disable=broad-except
        stderr('Caught {name} exception: {msg}'.format(name=e.__class__.__name__, msg=e))
        logger.exception('Failed to execute')
        exit_code = 99

    return exit_code


if __name__ == '__main__':
    sys.exit(main())



"""ClusterShell transport tests."""
# pylint: disable=invalid-name,no-member,protected-access,attribute-defined-outside-init
from unittest import mock

import pytest

from cumin import CuminError, nodeset
from cumin.transports import BaseWorker, clustershell, Command, ProgressBars, State, Target, WorkerError


def test_node_class_instantiation():
    """Default values should be set when a Node instance is created."""
    node = clustershell.Node('name', [Command('command1'), Command('command2')])
    assert node.running_command_index == -1
    assert isinstance(node.state, State)


@mock.patch('cumin.transports.clustershell.Task.task_self')
def test_worker_class(task_self):
    """An instance of worker_class should be an instance of BaseWorker."""
    worker = clustershell.worker_class({}, Target(nodeset('node1')))
    assert isinstance(worker, BaseWorker)
    task_self.assert_called_once_with()


class TestClusterShellWorker:
    """ClusterShell backend worker test class."""

    @mock.patch('cumin.transports.clustershell.Task.task_self')
    def setup_method(self, _, task_self):  # pylint: disable=arguments-differ
        """Initialize default properties and instances."""
        self.config = {
            'clustershell': {
                'ssh_options': ['-o StrictHostKeyChecking=no', '-o BatchMode=yes'],
                'fanout': 3}}

        self.target = Target(nodeset('node[1-2]'))
        self.worker = clustershell.worker_class(self.config, self.target)
        self.commands = [Command('command1'), Command('command2', ok_codes=[0, 100], timeout=5)]
        self.task_self = task_self
        # Mock default handlers
        clustershell.DEFAULT_HANDLERS = {
            'sync': mock.MagicMock(spec_set=clustershell.SyncEventHandler),
            'async': mock.MagicMock(spec_set=clustershell.AsyncEventHandler)}

        # Initialize the worker
        self.worker.commands = self.commands

    @mock.patch('cumin.transports.clustershell.Task.task_self')
    def test_instantiation(self, task_self):
        """An instance of ClusterShellWorker should be an instance of BaseWorker and initialize ClusterShell."""
        worker = clustershell.ClusterShellWorker(self.config, self.target)
        assert isinstance(worker, BaseWorker)
        task_self.assert_called_once_with()
        worker.task.set_info.assert_has_calls(
            [mock.call('fanout', 3),
             mock.call('ssh_options', '-o StrictHostKeyChecking=no -o BatchMode=yes')], any_order=True)

    def test_execute_default_sync_handler(self):
        """Calling execute() in sync mode without event handler should use the default sync event handler."""
        self.worker.handler = 'sync'
        self.worker.execute()
        args, kwargs = self.worker.task.shell.call_args
        assert args == ('command1',)
        assert kwargs['nodes'] == self.target.first_batch
        assert kwargs['handler'] == self.worker._handler_instance
        assert clustershell.DEFAULT_HANDLERS['sync'].called

    def test_execute_default_async_handler(self):
        """Calling execute() in async mode without event handler should use the default async event handler."""
        self.worker.handler = 'async'
        self.worker.execute()
        args, kwargs = self.worker.task.shell.call_args
        assert args == ('command1',)
        assert kwargs['nodes'] == self.target.first_batch
        assert kwargs['handler'] == self.worker._handler_instance
        assert clustershell.DEFAULT_HANDLERS['async'].called

    def test_execute_timeout(self):
        """Calling execute() and let the global timeout expire should call on_timeout."""
        self.worker.task.run = mock.Mock(side_effect=clustershell.Task.TimeoutError)
        self.worker.handler = 'sync'
        self.worker.execute()
        self.worker._handler_instance.on_timeout.assert_called_once_with(self.worker.task)

    def test_execute_custom_handler(self):
        """Calling execute() using a custom handler should call ClusterShell task with the custom event handler."""
        self.worker.handler = ConcreteBaseEventHandler
        self.worker.execute()
        assert isinstance(self.worker._handler_instance, ConcreteBaseEventHandler)
        args, kwargs = self.worker.task.shell.call_args
        assert args == ('command1',)
        assert kwargs['nodes'] == self.target.first_batch
        assert kwargs['handler'] == self.worker._handler_instance

    def test_execute_no_commands(self):
        """Calling execute() without commands should raise WorkerError."""
        self.worker.handler = ConcreteBaseEventHandler
        self.worker.commands = None
        with pytest.raises(WorkerError, match=r'No commands provided\.'):
            self.worker.execute()
        assert not self.worker.task.shell.called

    def test_execute_one_command_no_mode(self):
        """Calling execute() with only one command without mode should raise WorkerError."""
        self.worker.commands = [self.commands[0]]
        with pytest.raises(WorkerError, match=r'An EventHandler is mandatory\.'):
            self.worker.execute()
        assert not self.worker.task.shell.called

    def test_execute_wrong_mode(self):
        """Calling execute() without setting the mode with multiple commands should raise CuminError."""
        with pytest.raises(CuminError, match=r'An EventHandler is mandatory\.'):
            self.worker.execute()

    def test_execute_batch_size(self):
        """Calling execute() with a batch_size specified should run in batches."""
        self.worker.commands = [self.commands[0]]
        self.worker.handler = 'sync'
        self.worker.batch_size = 1
        self.worker.execute()
        args, kwargs = self.worker.task.shell.call_args
        assert args == ('command1',)
        assert kwargs['nodes'] == self.target.first_batch
        assert kwargs['handler'] == self.worker._handler_instance

    def test_get_results(self):
        """Calling get_results() should call ClusterShell iter_buffers with the right parameters."""
        self.worker.task.iter_buffers = TestClusterShellWorker.iter_buffers
        self.worker.handler = 'async'
        self.worker.execute()
        nodes = None
        output = None
        for nodes, output in self.worker.get_results():
            pass
        assert str(nodes) == 'node[90-92]'
        assert output == 'output 9'

    def test_handler_getter(self):
        """Access to handler property should return the handler class or None."""
        assert self.worker.handler is None
        self.worker.handler = 'sync'
        assert self.worker._handler == clustershell.DEFAULT_HANDLERS['sync']

    def test_handler_setter_invalid(self):
        """Raise WorkerError if trying to set it to an invalid class or value."""
        class InvalidClass:
            """Invalid class."""

        with pytest.raises(WorkerError, match='handler must be one of'):
            self.worker.handler = 'invalid-handler'

        with pytest.raises(WorkerError, match='handler must be one of'):
            self.worker.handler = InvalidClass

    def test_handler_setter_default_sync(self):
        """Should set the handler to the default handler for the sync mode."""
        self.worker.handler = 'sync'
        assert self.worker._handler == clustershell.DEFAULT_HANDLERS['sync']

    def test_handler_setter_default_async(self):
        """Should set the handler to the default handler for the async mode."""
        self.worker.handler = 'async'
        assert self.worker._handler == clustershell.DEFAULT_HANDLERS['async']

    def test_handler_setter_custom(self):
        """Should set the handler to the given custom class that inherit from BaseEventHandler."""
        self.worker.handler = ConcreteBaseEventHandler
        assert self.worker._handler == ConcreteBaseEventHandler

    @staticmethod
    def iter_buffers():
        """A generator to simulate the buffer iteration of ClusterShell objects."""
        for i in range(10):
            yield 'output {}'.format(i), ['node{}0'.format(i), 'node{}1'.format(i), 'node{}2'.format(i)]


class TestBaseEventHandler:
    """BaseEventHandler test class."""

    def setup_method(self, *args):  # pylint: disable=arguments-differ
        """Initialize default properties and instances."""
        self.target = Target(nodeset('node[1-2]'))
        self.commands = [Command('command1', ok_codes=[0, 100]), Command('command2', timeout=5)]
        self.worker = mock.MagicMock()
        self.worker.current_node = 'node1'
        self.worker.command = 'command1'
        self.worker.nodes = self.target.hosts
        self.handler = None
        self.args = args

    def test_close(self):
        """Calling close should raise NotImplementedError."""
        self.handler = clustershell.BaseEventHandler(self.target, self.commands)
        with pytest.raises(NotImplementedError):
            self.handler.close(self.worker)


class ConcreteBaseEventHandler(clustershell.BaseEventHandler):
    """Concrete implementation of a BaseEventHandler."""

    def __init__(self, nodes, commands, **kwargs):
        """Initialize progress bars."""
        super().__init__(nodes, commands, **kwargs)
        self.progress = mock.Mock(spec_set=ProgressBars)

    def close(self, task):
        """Required by the BaseEventHandler class."""


class TestConcreteBaseEventHandler(TestBaseEventHandler):
    """ConcreteBaseEventHandler test class."""

    @mock.patch('cumin.transports.clustershell.tqdm')
    def setup_method(self, _, tqdm):  # pylint: disable=arguments-differ
        """Initialize default properties and instances."""
        super().setup_method()
        self.handler = ConcreteBaseEventHandler(self.target, self.commands)
        self.worker.eh = self.handler
        assert not tqdm.write.called

    def test_instantiation(self):
        """An instance of ConcreteBaseEventHandler should be an instance of BaseEventHandler."""
        assert sorted(self.handler.nodes.keys()) == list(self.target.hosts)

    @mock.patch('cumin.transports.clustershell.tqdm')
    def test_on_timeout(self, tqdm):
        """Calling on_timeout() should update the fail progress bar."""
        for node in self.target.hosts:
            self.worker.current_node = node
            self.handler.ev_pickup(self.worker, node)
            self.handler.ev_close(self.worker, True)
        self.worker.task.num_timeout.return_value = 1
        self.worker.task.iter_keys_timeout.return_value = [self.target.hosts[0]]

        assert not self.handler.global_timedout
        self.handler.on_timeout(self.worker.task)
        assert self.handler.progress.update_failed.called
        assert self.handler.global_timedout
        assert tqdm.write.called

    def test_ev_pickup(self):
        """Calling ev_pickup() should set the state of the current node to running."""
        for node in self.target.hosts:
            self.handler.ev_pickup(self.worker, node)
        running_nodes = [node for node in self.worker.eh.nodes.values() if node.state.is_running]
        assert running_nodes == list(self.worker.eh.nodes.values())

    @mock.patch('cumin.transports.clustershell.tqdm')
    def test_ev_read_many_hosts(self, tqdm):
        """Calling ev_read() should not print the worker message if matching multiple hosts."""
        for node in self.target.hosts:
            self.handler.ev_read(self.worker, node, self.worker.SNAME_STDOUT, 'Node output')
        assert not tqdm.write.called

    @mock.patch('cumin.transports.clustershell.tqdm')
    def test_ev_read_single_host(self, tqdm):
        """Calling ev_read() should print the worker message if matching a single host."""
        self.target = Target(nodeset('node1'))
        self.handler = ConcreteBaseEventHandler(self.target, self.commands)

        output = b'node1 output'
        self.worker.nodes = self.target.hosts
        self.handler.ev_read(self.worker, self.target.hosts[0], self.worker.SNAME_STDOUT, output)
        assert tqdm.write.call_args[0][0] == output.decode()

    def test_ev_close(self):
        """Calling ev_close() should increase the counters for the timed out hosts."""
        for node in self.target.hosts:
            self.handler.ev_pickup(self.worker, node)

        assert self.handler.counters['timeout'] == 0
        self.worker.task.num_timeout.return_value = 2
        self.handler.ev_close(self.worker, True)
        assert self.handler.counters['timeout'] == 2


class TestSyncEventHandler(TestBaseEventHandler):
    """SyncEventHandler test class."""

    @mock.patch('cumin.transports.clustershell.logging')
    @mock.patch('cumin.transports.clustershell.tqdm')
    def setup_method(self, _, tqdm, logger):  # pylint: disable=arguments-differ
        """Initialize default properties and instances."""
        super().setup_method()
        self.handler = clustershell.SyncEventHandler(self.target, self.commands, success_threshold=1)
        self.handler.progress = mock.Mock(spec_set=ProgressBars)
        self.worker.eh = self.handler
        self.logger = logger
        assert not tqdm.write.called

    def test_instantiation(self):
        """An instance of SyncEventHandler should be an instance of BaseEventHandler."""
        assert isinstance(self.handler, clustershell.BaseEventHandler)

    def test_start_command_no_schedule(self):
        """Calling start_command() should reset the success counter and initialize the progress bars."""
        self.handler.start_command()
        assert self.handler.progress.init.called
        assert self.handler.counters['success'] == 0

    @mock.patch('cumin.transports.clustershell.Task.task_self')
    def test_start_command_schedule(self, task_self):
        """Calling start_command() with schedule should also change the state of the first batch nodes."""
        # Reset the state of nodes to pending
        for node in self.handler.nodes.values():
            node.state.update(clustershell.State.running)
            node.state.update(clustershell.State.success)
            node.state.update(clustershell.State.pending)

        self.handler.start_command(schedule=True)
        assert self.handler.progress.init.called
        assert self.handler.counters['success'] == 0
        scheduled_nodes = sorted(node.name for node in self.handler.nodes.values() if node.state.is_scheduled)
        assert scheduled_nodes == sorted(['node1', 'node2'])
        assert task_self.called

    @mock.patch('cumin.transports.clustershell.tqdm')
    def test_end_command(self, tqdm):
        """Calling end_command() should wrap up the command execution."""
        assert not self.handler.end_command()
        self.handler.counters['success'] = 2
        assert self.handler.end_command()
        self.handler.success_threshold = 0.5
        self.handler.counters['success'] = 1
        assert self.handler.end_command()
        self.handler.current_command_index = 1
        assert not self.handler.end_command()
        assert tqdm.write.called

    @mock.patch('cumin.transports.clustershell.tqdm')
    def test_on_timeout(self, tqdm):
        """Calling on_timeout() should call end_command()."""
        self.worker.task.num_timeout.return_value = 0
        self.worker.task.iter_keys_timeout.return_value = []
        self.handler.on_timeout(self.worker.task)
        assert tqdm.write.called

    def test_ev_timer(self):
        """Calling ev_timer() should schedule the execution of the next node/command."""
        # TODO: improve testing of ev_timer
        self.handler.ev_timer(mock.Mock())

    @mock.patch('cumin.transports.clustershell.Task.Task.timer')
    def test_ev_hup_ok(self, timer):
        """Calling ev_hup with a worker that has exit status zero should update the success progress bar."""
        self.handler.ev_pickup(self.worker, self.worker.current_node)
        self.handler.ev_hup(self.worker, self.worker.current_node, 100)
        assert self.handler.progress.update_success.called
        assert not timer.called
        assert self.handler.nodes[self.worker.current_node].state.is_success

    @mock.patch('cumin.transports.clustershell.Task.Task.timer')
    def test_ev_hup_ko(self, timer):
        """Calling ev_hup with a worker that has exit status non-zero should update the failed progress bar."""
        self.handler.ev_pickup(self.worker, self.worker.current_node)
        self.handler.ev_hup(self.worker, self.worker.current_node, 1)
        assert self.handler.progress.update_failed.called
        assert not timer.called
        assert self.handler.nodes[self.worker.current_node].state.is_failed

    @mock.patch('cumin.transports.clustershell.tqdm')
    def test_close(self, tqdm):  # pylint: disable=arguments-differ
        """Calling close should print the report when needed."""
        self.handler.current_command_index = 2
        self.handler.close(self.worker)
        assert tqdm.write.called


class TestAsyncEventHandler(TestBaseEventHandler):
    """AsyncEventHandler test class."""

    @mock.patch('cumin.transports.clustershell.ProgressBars')
    @mock.patch('cumin.transports.clustershell.logging')
    @mock.patch('cumin.transports.clustershell.tqdm')
    def setup_method(self, _, tqdm, logger, progress):  # pylint: disable=arguments-differ,unused-argument
        """Initialize default properties and instances."""
        super().setup_method()
        self.handler = clustershell.AsyncEventHandler(self.target, self.commands)
        self.worker.eh = self.handler
        self.logger = logger
        assert not tqdm.write.called

    def test_instantiation(self):
        """An instance of AsyncEventHandler should be an instance of BaseEventHandler and initialize progress bars."""
        assert isinstance(self.handler, clustershell.BaseEventHandler)
        assert self.handler.progress.init.called

    def test_ev_hup_ok(self):
        """Calling ev_hup with a worker that has zero exit status should enqueue the next command."""
        self.handler.ev_pickup(self.worker, self.worker.current_node)
        self.handler.ev_hup(self.worker, self.worker.current_node, 0)
        self.worker.task.shell.assert_called_once_with(
            'command2', handler=self.handler, timeout=5, stdin=False, nodes=nodeset(self.worker.current_node))

        # Calling it again
        self.worker.command = 'command2'
        self.handler.ev_pickup(self.worker, self.worker.current_node)
        self.handler.ev_hup(self.worker, self.worker.current_node, 0)
        assert self.handler.counters['success'] == 1
        assert self.handler.progress.update_success.called

    def test_ev_hup_ko(self):
        """Calling ev_hup with a worker that has non-zero exit status should not enqueue the next command."""
        self.handler.ev_pickup(self.worker, self.worker.current_node)
        self.handler.ev_hup(self.worker, self.worker.current_node, 1)
        assert self.handler.progress.update_failed.called

    def test_ev_timer(self):
        """Calling ev_timer() should schedule the execution of the next node/command."""
        # TODO: improve testing of ev_timer
        self.handler.ev_timer(mock.Mock())

    @mock.patch('cumin.transports.clustershell.tqdm')
    def test_close(self, tqdm):  # pylint: disable=arguments-differ
        """Calling close with a worker should close progress bars."""
        self.worker.task.iter_buffers = TestClusterShellWorker.iter_buffers
        self.worker.num_timeout.return_value = 0
        self.handler.close(self.worker)
        assert self.handler.progress.close.called
        assert tqdm.write.called



"""Test external backend module with missing GRAMMAR_PREFIX."""
from cumin.tests.unit.backends.external import ExternalBackendQuery


query_class = ExternalBackendQuery  # pylint: disable=invalid-name
"""Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""



"""Query handling tests."""
# pylint: disable=invalid-name

import pytest

from cumin import backends, nodeset
from cumin.query import Query


def test_execute_valid_global():
    """Executing a valid query should return the matching hosts."""
    query = Query({})
    hosts = query.execute('D{(host1 or host2) and host[1-5]}')
    assert hosts == nodeset('host[1-2]')


def test_execute_global_or():
    """Executing an 'or' between two queries should return the union of the hosts."""
    query = Query({})
    hosts = query.execute('D{host1} or D{host2}')
    assert hosts == nodeset('host[1-2]')


def test_execute_global_and():
    """Executing an 'and' between two queries should return the intersection of the hosts."""
    query = Query({})
    hosts = query.execute('D{host[1-5]} and D{host2}')
    assert hosts == nodeset('host2')


def test_execute_global_and_not():
    """Executing an 'and not' between two queries should return the difference of the hosts."""
    query = Query({})
    hosts = query.execute('D{host[1-5]} and not D{host2}')
    assert hosts == nodeset('host[1,3-5]')


def test_execute_global_xor():
    """Executing a 'xor' between two queries should return all the hosts that are in only in one of the queries."""
    query = Query({})
    hosts = query.execute('D{host[1-5]} xor D{host[3-7]}')
    assert hosts == nodeset('host[1-2,6-7]')


def test_execute_valid_global_with_aliases():
    """Executing a valid query with aliases should return the matching hosts."""
    query = Query({'aliases': {'group1': 'D{host1 or host2}'}})
    hosts = query.execute('A:group1')
    assert hosts == nodeset('host[1-2]')


def test_execute_valid_global_with_nested_aliases():
    """Executing a valid query with nested aliases should return the matching hosts."""
    query = Query({
        'aliases': {
            'group1': 'D{host1 or host2}',
            'group2': 'D{host3 or host4}',
            'all': 'A:group1 or A:group2',
        }})
    hosts = query.execute('A:all')
    assert hosts == nodeset('host[1-4]')


def test_execute_missing_alias():
    """Executing a valid query with a missing alias should raise InvalidQueryError."""
    query = Query({})
    with pytest.raises(backends.InvalidQueryError, match='Unable to find alias replacement for'):
        query.execute('A:non_existent_group')

    query = Query({'aliases': {}})
    with pytest.raises(backends.InvalidQueryError, match='Unable to find alias replacement for'):
        query.execute('A:non_existent_group')


def test_execute_invalid_global():
    """Executing a query with an invalid syntax should raise InvalidQueryError."""
    query = Query({})
    with pytest.raises(backends.InvalidQueryError, match='with the global grammar'):
        query.execute('invalid syntax')


def test_execute_subgroup():
    """Executing a query with a single subgroup should return the matching hosts."""
    query = Query({})
    hosts = query.execute('(D{host1})')
    assert hosts == nodeset('host1')


def test_execute_subgroups():
    """Executing a query with multiple subgroups should return the matching hosts."""
    query = Query({})
    hosts = query.execute('(D{host1} or D{host2}) and not (D{host1})')
    assert hosts == nodeset('host2')


def test_execute_missing_default_backend():
    """Executing a valid query with a missing default backend should raise InvalidQueryError."""
    query = Query({'default_backend': 'non_existent_backend'})
    with pytest.raises(backends.InvalidQueryError, match='is not registered'):
        query.execute('any_query')


def test_execute_valid_default_backend():
    """Executing a default backend valid query should return the matching hosts."""
    query = Query({'default_backend': 'direct'})
    hosts = query.execute('host1 or host2')
    assert hosts == nodeset('host[1-2]')


def test_execute_invalid_default_valid_global():
    """Executing a global grammar valid query in presence of a default backend should return the matching hosts."""
    query = Query({'default_backend': 'direct'})
    hosts = query.execute('D{host1 or host2}')
    assert hosts == nodeset('host[1-2]')


def test_execute_invalid_default_invalid_global():
    """Executing a query invalid for the default backend and the global grammar should raise InvalidQueryError."""
    query = Query({'default_backend': 'direct'})
    with pytest.raises(backends.InvalidQueryError, match='neither with the default backen'):
        query.execute('invalid syntax')


def test_execute_complex_global():
    """Executing a valid complex query should return the matching hosts."""
    query = Query({})
    hosts = query.execute(
        '(D{(host1 or host2) and host[1-5]}) or ((D{host[100-150]} and not D{host1[20-30]}) and D{host1[01,15,30]})')
    assert hosts == nodeset('host[1-2,101,115]')



"""Automation and orchestration framework written in Python."""
import logging
import os

import yaml

from ClusterShell.NodeSet import NodeSet, RESOLVER_NOGROUP
from pkg_resources import DistributionNotFound, get_distribution


try:
    __version__ = get_distribution(__name__).version
    """:py:class:`str`: the version of the current Cumin module."""
except DistributionNotFound:  # pragma: no cover - this should never happen during tests
    pass  # package is not installed


class CuminError(Exception):
    """Base Exception class for all Cumin's custom Exceptions."""


##############################################################################
# Add a custom log level TRACE to logging for development debugging

LOGGING_TRACE_LEVEL_NUMBER = 8
LOGGING_TRACE_LEVEL_NAME = 'TRACE'


# Fail if the custom logging slot is already in use with a different name or
# Access to a private property of logging was preferred over matching the default string returned by
# logging.getLevelName() for unused custom slots.
if (LOGGING_TRACE_LEVEL_NUMBER in logging._levelToName  # pylint: disable=protected-access
        and LOGGING_TRACE_LEVEL_NAME not in logging._nameToLevel):  # pylint: disable=protected-access
    raise CuminError("Unable to set custom logging for trace, logging level {level} is alredy set for '{name}'.".format(
        level=LOGGING_TRACE_LEVEL_NUMBER, name=logging.getLevelName(LOGGING_TRACE_LEVEL_NUMBER)))


def trace(self, msg, *args, **kwargs):
    """Additional logging level for development debugging.

    :Parameters:
        according to :py:class:`logging.Logger` interface for log levels.

    """
    if self.isEnabledFor(LOGGING_TRACE_LEVEL_NUMBER):
        self._log(LOGGING_TRACE_LEVEL_NUMBER, msg, args, **kwargs)  # pragma: no cover, pylint: disable=protected-access


# Install the trace method and it's logging level if not already present
if LOGGING_TRACE_LEVEL_NAME not in logging._nameToLevel:  # pylint: disable=protected-access
    logging.addLevelName(LOGGING_TRACE_LEVEL_NUMBER, LOGGING_TRACE_LEVEL_NAME)
if not hasattr(logging.Logger, 'trace'):
    logging.Logger.trace = trace
##############################################################################


class Config(dict):
    """Singleton-like dictionary class to load the configuration from a given path only once."""

    _instances = {}  # Keep track of different loaded configurations

    def __new__(cls, config='/etc/cumin/config.yaml'):
        """Load the given configuration if not already loaded and return it.

        Called by Python's data model for each new instantiation of the class.

        Arguments:
            config (str, optional): path to the configuration file to load.

        Returns:
            dict: the configuration dictionary.

        Examples:
            >>> import cumin
            >>> config = cumin.Config()

        """
        if config not in cls._instances:
            cls._instances[config] = parse_config(config)
            alias_file = os.path.join(os.path.dirname(config), 'aliases.yaml')
            if os.path.isfile(alias_file):  # Load the aliases only if present
                cls._instances[config]['aliases'] = parse_config(alias_file)

        return cls._instances[config]


def parse_config(config_file):
    """Parse the YAML configuration file.

    Arguments:
        config_file (str): the path of the configuration file to load.

    Returns:
        dict: the configuration dictionary.

    Raises:
        CuminError: if unable to read or parse the configuration.

    """
    try:
        with open(config_file, 'r') as f:
            config = yaml.safe_load(f)
    except IOError as e:
        raise CuminError('Unable to read configuration file: {message}'.format(message=e))
    except yaml.parser.ParserError as e:
        raise CuminError("Unable to parse configuration file '{config}':\n{message}".format(
            config=config_file, message=e))

    if config is None:
        config = {}

    return config


def nodeset(nodes=None):
    """Instantiate a ClusterShell NodeSet with the resolver defaulting to :py:const:`RESOLVER_NOGROUP`.

    This allow to avoid any conflict with Cumin grammars.

    Returns:
        ClusterShell.NodeSet.NodeSet: the instantiated NodeSet.

    See Also:
        https://github.com/cea-hpc/clustershell/issues/368

    """
    return NodeSet(nodes=nodes, resolver=RESOLVER_NOGROUP)


def nodeset_fromlist(nodelist):
    """Instantiate a ClusterShell NodeSet from a list with the resolver defaulting to :py:const:`RESOLVER_NOGROUP`.

    This allow to avoid any conflict with Cumin grammars.

    Returns:
        ClusterShell.NodeSet.NodeSet: the instantiated NodeSet.

    See Also:
        https://github.com/cea-hpc/clustershell/issues/368

    """
    return NodeSet.fromlist(nodelist, resolver=RESOLVER_NOGROUP)



#!/usr/bin/env python
"""Package configuration."""
import os

from setuptools import find_packages, setup


with open('README.rst', 'r') as readme:
    long_description = readme.read()

# Required dependencies
install_requires = [
    'clustershell>=1.8.1,<=1.9',
    'pyparsing>=2.2.0,<=2.3',
    'pyyaml>=3.13',
    'requests>=2.21.0',
    'tqdm>=4.19.4,<=4.24.0'
]

# Extra dependencies
extras_require = {
    # Optional dependencies for additional features
    'with-openstack': [
        'keystoneauth1>=3.10.0',
        'python-keystoneclient>=3.17.0',
        'python-novaclient>=11.0.0',
    ],

    # Test dependencies
    'tests': [
        'bandit>=1.5.1',
        'flake8>=3.6.0',
        'flake8-import-order>=0.18.1',
        # Temporary upper constraint until https://github.com/PyCQA/prospector/issues/389 is fixed
        'prospector[with_everything]>=1.1.7,<=1.2.0',
        'pytest-cov>=2.6.0',
        'pytest-xdist>=1.26.1',
        'pytest>=3.10.1',
        'requests-mock>=1.5.2',
        'sphinx_rtd_theme>=0.4.3',
        'sphinx-argparse>=0.2.2',
        'Sphinx>=1.8.4',
    ],
}

# Copy tests requirements to test only base dependencies
extras_require['tests-base'] = extras_require['tests'][:]
# Add optional dependencies to the tests ones
extras_require['tests'].extend(extras_require['with-openstack'])

# Generate minimum dependencies
extras_require['tests-min'] = [dep.replace('>=', '==') for dep in extras_require['tests']]
if os.getenv('CUMIN_MIN_DEPS', False):
    install_requires = [dep.split(',')[0].replace('>=', '==') for dep in install_requires]

setup_requires = [
    'pytest-runner>=2.11.1',
    'setuptools_scm>=3.2.0',
]

setup(
    author='Riccardo Coccioli',
    author_email='rcoccioli@wikimedia.org',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: BSD',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Clustering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Systems Administration',
    ],
    data_files=[('share/doc/cumin/examples/', ['doc/examples/config.yaml', 'doc/examples/aliases.yaml'])],
    description='Automation and orchestration framework and CLI written in Python',
    entry_points={
        'console_scripts': [
            'cumin = cumin.cli:main',
        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    keywords=['cumin', 'automation', 'orchestration'],
    license='GPLv3+',
    long_description=long_description,
    name='cumin',
    packages=find_packages(exclude=['*.tests', '*.tests.*']),
    platforms=['GNU/Linux', 'BSD', 'MacOSX'],
    setup_requires=setup_requires,
    url='https://github.com/wikimedia/cumin',
    use_scm_version=True,
    zip_safe=False,
)



"""Direct backend tests."""

from cumin import nodeset
from cumin.backends import BaseQuery, direct


def test_direct_query_class():
    """An instance of query_class should be an instance of BaseQuery."""
    query = direct.query_class({})
    assert isinstance(query, BaseQuery)


class TestDirectQuery:
    """Direct backend query test class."""

    def setup_method(self, _):
        """Setup an instance of DirectQuery for each test."""
        self.query = direct.DirectQuery({})  # pylint: disable=attribute-defined-outside-init

    def test_instantiation(self):
        """An instance of DirectQuery should be an instance of BaseQuery."""
        assert isinstance(self.query, BaseQuery)
        assert self.query.config == {}

    def test_execute(self):
        """Calling execute() should return the list of hosts."""
        assert self.query.execute('host1 or host2') == nodeset('host[1-2]')
        assert self.query.execute('host1 and host2') == nodeset()
        assert self.query.execute('host1 and not host2') == nodeset('host1')
        assert self.query.execute('host[1-5] xor host[3-7]') == nodeset('host[1-2,6-7]')
        assert self.query.execute('host1 or (host[10-20] and not host15)') == nodeset('host[1,10-14,16-20]')
        assert self.query.execute('(host1 or host[2-3]) and not (host[3-9] or host2)') == nodeset('host1')



# -*- coding: utf-8 -*-
"""Sphynx configuration."""
#
# Cumin documentation build configuration file, created by
# sphinx-quickstart on Sat Sep 30 13:33:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import sphinx_rtd_theme

from pkg_resources import get_distribution
from sphinx import __version__ as sphinx_version

# Adjust path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir)))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinxarg.ext',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Cumin'
title = u'{project} Documentation'.format(project=project)
copyright = u"2017-2020, Riccardo Coccioli <rcoccioli@wikimedia.org>, Wikimedia Foundation, Inc."
author = u'Riccardo Coccioli'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = get_distribution('cumin').version
# The short X.Y version.
version = release


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
sphinx_version_parts = [int(i) for i in sphinx_version.split('.')]
if sphinx_version_parts[0] == 1 and sphinx_version_parts[1] < 6:
    html_use_smartypants = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Cumindoc'


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('cli', 'cumin', 'Automation and orchestration framework written in Python', [author], 1),
]


# -- Options for intersphinx ---------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'requests': ('https://requests.readthedocs.io/en/master/', None),
    'ClusterShell': ('https://clustershell.readthedocs.io/en/v1.8.1/', None),
    'keystoneauth1': ('https://docs.openstack.org/keystoneauth/latest', None),
    'novaclient': ('https://docs.openstack.org/python-novaclient/latest/', None),
    'pyparsing': ('https://pythonhosted.org/pyparsing/', 'pyparsing.inv'),
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_use_keyword = True

# Autodoc settings
autodoc_default_options = {
    # Using None as value instead of True to support the version of Sphinx used in Buster
    'members': None,
    'member-order': 'bysource',
    'private-members': None,
    'show-inheritance': None,
}
autoclass_content = 'both'


# -- Helper functions -----------------------------------------------------
def filter_namedtuple_docstrings(app, what, name, obj, options, lines):
    """Fix the automatically generated docstrings for namedtuples classes."""
    if what == 'property' and len(lines) == 1 and lines[0].startswith('Alias for field number'):
        del lines[:]


# Keep track of documented classes to avoid annotating both class and __init__.
# Necessary when using autoclass_content 'both' and add_abstract_annotations().
_cumin_documented_classes = set()


def add_abstract_annotations(app, what, name, obj, options, lines):
    """Workaround to add an abstract annotation for ABC abstract classes."""
    if what == 'class' and len(getattr(obj, '__abstractmethods__', [])) > 0 and name not in _cumin_documented_classes:
        lines.insert(0, '``abstract``')
        _cumin_documented_classes.add(name)


def setup(app):
    """Register the helper functions."""
    app.connect('autodoc-process-docstring', filter_namedtuple_docstrings)
    app.connect('autodoc-process-docstring', add_abstract_annotations)
    app.add_stylesheet('theme_overrides.css')  # override wide tables in RTD theme



"""Tests utils."""

import logging
import os

logging.basicConfig(level=logging.DEBUG)
_TESTS_BASE_PATH = os.path.realpath(os.path.dirname(__file__))


def get_fixture(path, as_string=False):
    """Return the content of a fixture file.

    Arguments:
        path: the relative path to the test's fixture directory to be opened.
        as_string: return the content as a multiline string instead of a list of lines [optional, default: False]

    """
    with open(get_fixture_path(path)) as f:
        if as_string:
            content = f.read()
        else:
            content = f.readlines()

    return content


def get_fixture_path(path):
    """Return the absolute path of the given fixture.

    Arguments:
        path: the relative path to the test's fixture directory.

    """
    return os.path.join(_TESTS_BASE_PATH, 'fixtures', path)



"""Known hosts backend tests."""
import os

import pytest

from ClusterShell.NodeSet import NodeSet, RESOLVER_NOGROUP

from cumin.backends import BaseQuery
from cumin.backends.knownhosts import KnownHostsLineError, KnownHostsQuery, KnownHostsSkippedLineError, query_class
from cumin.tests import get_fixture_path


def test_knownhosts_query_class():
    """An instance of query_class should be an instance of BaseQuery."""
    query = query_class({})
    assert isinstance(query, BaseQuery)


class TestKnownhostsQuery:
    """Knownhosts backend query test class."""

    def setup_method(self, _):
        """Set up an instance of KnownHostsQuery for each test."""
        # pylint: disable=attribute-defined-outside-init
        self.query = KnownHostsQuery({
            'knownhosts': {'files': [
                get_fixture_path(os.path.join('backends', 'knownhosts.txt')),
                get_fixture_path(os.path.join('backends', 'knownhosts_man.txt')),
            ]}})
        self.no_query = KnownHostsQuery({})
        self.no_hosts = NodeSet(resolver=RESOLVER_NOGROUP)
        self.domain_hosts = NodeSet('host[1,4-5,7-8,13-14].domain', resolver=RESOLVER_NOGROUP)
        self.all_hosts = self.domain_hosts | NodeSet('closenet,cvs.example.net', resolver=RESOLVER_NOGROUP)

    def test_instantiation(self):
        """An instance of KnownHostsQuery should be an instance of BaseQuery."""
        assert isinstance(self.query, BaseQuery)
        assert 'knownhosts' in self.query.config

    def test_execute(self):
        """Calling execute() with one host should return it."""
        assert self.query.execute('host1.domain') == NodeSet('host1.domain', resolver=RESOLVER_NOGROUP)

    def test_execute_non_existent(self):
        """Calling execute() with one host that doens't exists should return no hosts."""
        assert self.query.execute('nohost1.domain') == self.no_hosts

    def test_execute_or(self):
        """Calling execute() with two hosts in 'or' should return both hosts."""
        expected = NodeSet('host[1,4].domain', resolver=RESOLVER_NOGROUP)
        assert self.query.execute('host1.domain or host4.domain') == expected

    def test_execute_and(self):
        """Calling execute() with two hosts in 'and' should return no hosts."""
        assert self.query.execute('host1.domain and host2.domain') == self.no_hosts

    def test_execute_and_not(self):
        """Calling execute() with two hosts with 'and not' should return the first host."""
        expected = NodeSet('host1.domain', resolver=RESOLVER_NOGROUP)
        assert self.query.execute('host1.domain and not host2.domain') == expected

    def test_execute_xor(self):
        """Calling execute() with two host groups with 'xor' should return the hosts that are not in both groups."""
        expected = NodeSet('host[1,7-8].domain', resolver=RESOLVER_NOGROUP)
        assert self.query.execute('host[1-8].domain xor host[4-6].domain') == expected

    def test_execute_complex(self):
        """Calling execute() with a complex query should return the matching hosts."""
        expected = NodeSet('host[1,5,8].domain', resolver=RESOLVER_NOGROUP)
        assert self.query.execute('host1.domain or (host[5-9].domain and not host7.domain)') == expected

        expected = NodeSet('host1.domain', resolver=RESOLVER_NOGROUP)
        assert self.query.execute(
            '(host1.domain or host[2-5].domain) and not (host[3-9].domain or host2.domain)') == expected

    def test_execute_all(self):
        """Calling execute() with broader matching should return all hosts."""
        assert self.query.execute('*') == self.all_hosts
        assert self.query.execute('host[1-100].domain') == self.domain_hosts
        assert self.query.execute('host[1-100].domai?') == self.domain_hosts
        assert self.query.execute('host[1-100].*') == self.domain_hosts

    def test_execute_no_hosts(self):
        """Calling execute() without any known hosts to load should return no hosts."""
        assert self.no_query.execute('host1.domain') == self.no_hosts
        assert self.no_query.execute('*') == self.no_hosts


def test_parse_line_empty():
    """Empty lines should raise KnownHostsSkippedLineError."""
    with pytest.raises(KnownHostsSkippedLineError, match='empty line'):
        KnownHostsQuery.parse_known_hosts_line('')
    with pytest.raises(KnownHostsSkippedLineError, match='empty line'):
        KnownHostsQuery.parse_known_hosts_line('\n')


def test_parse_line_comment():
    """Comment lines should raise KnownHostsSkippedLineError."""
    with pytest.raises(KnownHostsSkippedLineError, match='comment'):
        KnownHostsQuery.parse_known_hosts_line('# comment')


def test_parse_line_hashed():
    """Hashed lines should raise KnownHostsSkippedLineError."""
    with pytest.raises(KnownHostsSkippedLineError, match='hashed'):
        KnownHostsQuery.parse_known_hosts_line('|1|HaSh=|HaSh= ecdsa-sha2-nistp256 AAAA...=')


def test_parse_line_no_fields():
    """Lines without enough fields should raise KnownHostsLineError."""
    with pytest.raises(KnownHostsLineError, match='not enough fields'):
        KnownHostsQuery.parse_known_hosts_line('host1 ssh-rsa')


def test_parse_line_no_fields_mark():
    """Lines with a marker but without enough fields should raise KnownHostsLineError."""
    with pytest.raises(KnownHostsLineError, match='not enough fields'):
        KnownHostsQuery.parse_known_hosts_line('@marker host1 ssh-rsa')


def test_parse_line_revoked():
    """Lines with a revoked marker should raise KnownHostsSkippedLineError."""
    with pytest.raises(KnownHostsSkippedLineError, match='revoked'):
        KnownHostsQuery.parse_known_hosts_line('@revoked host1 ecdsa-sha2-nistp256 AAAA...=')


def test_parse_line_unknown_marker():
    """Lines with an unknown marker should raise KnownHostsLineError."""
    with pytest.raises(KnownHostsLineError, match='unknown marker'):
        KnownHostsQuery.parse_known_hosts_line('@marker host1 ecdsa-sha2-nistp256 AAAA...=')


def test_parse_line_ca():
    """Lines with a cert-authority marker should parse the hostnames."""
    expected = ({'host1'}, set())
    assert KnownHostsQuery.parse_known_hosts_line('@cert-authority host1 ecdsa-sha2-nistp256 AAAA...=') == expected


def test_parse_line():
    """With a standard line should parse the hostnames."""
    assert KnownHostsQuery.parse_known_hosts_line('host1 ecdsa-sha2-nistp256 AAAA...=') == ({'host1'}, set())


def test_parse_line_hosts_empty():
    """Empty line hosts should be skipped."""
    assert KnownHostsQuery.parse_line_hosts(',') == (set(), set())
    assert KnownHostsQuery.parse_line_hosts('host1,,') == ({'host1'}, set())


def test_parse_line_hosts_negated():
    """Negated line hosts should remove the negation."""
    assert KnownHostsQuery.parse_line_hosts('!host1') == ({'host1'}, set())
    expected = ({'host1', 'host2'}, set())
    assert KnownHostsQuery.parse_line_hosts('!host1,host2') == expected
    assert KnownHostsQuery.parse_line_hosts('host1,!host2') == expected
    assert KnownHostsQuery.parse_line_hosts('!host1,!host2') == expected


def test_parse_line_hosts_port():
    """Line hosts with custom ports should remove the additional syntax."""
    assert KnownHostsQuery.parse_line_hosts('[host1]:2222') == ({'host1'}, set())
    expected = ({'host1', 'host2'}, set())
    assert KnownHostsQuery.parse_line_hosts('[host1]:2222,host2') == expected
    assert KnownHostsQuery.parse_line_hosts('host1,[host2]:2222') == expected
    assert KnownHostsQuery.parse_line_hosts('[host1]:2222,[host2]:2222') == expected


def test_parse_line_hosts_neg_port():
    """Line hosts with custom ports and negated entries should remove the additional syntax."""
    assert KnownHostsQuery.parse_line_hosts('![host1]:2222') == ({'host1'}, set())
    expected = ({'host1', 'host2'}, set())
    assert KnownHostsQuery.parse_line_hosts('![host1]:2222,!host2') == expected
    assert KnownHostsQuery.parse_line_hosts('!host1,![host2]:2222') == expected
    assert KnownHostsQuery.parse_line_hosts('![host1]:2222,![host2]:2222') == expected


def test_parse_line_hosts_patterns():
    """Line hosts with patterns should skip the patterns entries."""
    assert KnownHostsQuery.parse_line_hosts('host?') == (set(), {'host?'})
    assert KnownHostsQuery.parse_line_hosts('host*') == (set(), {'host*'})
    assert KnownHostsQuery.parse_line_hosts('host?,host2') == ({'host2'}, {'host?'})
    assert KnownHostsQuery.parse_line_hosts('host*,host2') == ({'host2'}, {'host*'})
    assert KnownHostsQuery.parse_line_hosts('host*,host2,host?') == ({'host2'}, {'host?', 'host*'})


def test_parse_line_hosts_ips():
    """Line hosts with IPs should skip the IP entries."""
    assert KnownHostsQuery.parse_line_hosts('127.0.1.1') == (set(), {'127.0.1.1'})
    assert KnownHostsQuery.parse_line_hosts('fe80::1') == (set(), {'fe80::1'})
    assert KnownHostsQuery.parse_line_hosts('host1,127.0.1.1') == ({'host1'}, {'127.0.1.1'})
    assert KnownHostsQuery.parse_line_hosts('host1,fe80::1') == ({'host1'}, {'fe80::1'})
    assert KnownHostsQuery.parse_line_hosts('host1,127.0.1.1,fe80::1') == ({'host1'}, {'127.0.1.1', 'fe80::1'})



"""PuppetDB backend tests."""
# pylint: disable=invalid-name
from unittest import mock

import pytest

from requests.exceptions import HTTPError

from cumin import nodeset
from cumin.backends import BaseQuery, InvalidQueryError, puppetdb


def test_puppetdb_query_class():
    """An instance of query_class should be an instance of BaseQuery."""
    query = puppetdb.query_class({})
    assert isinstance(query, BaseQuery)


def _get_category_key_token(category='F', key='key1', operator='=', value='value1'):
    """Generate and return a category token string and it's expected dictionary of tokens when parsed."""
    expected = {'category': category, 'key': key, 'operator': operator, 'quoted': value}
    token = '{category}:{key} {operator} {quoted}'.format(**expected)
    return token, expected


def test_single_category_key_token():
    """A valid single token with a category that has key is properly parsed and interpreted."""
    token, expected = _get_category_key_token()
    parsed = puppetdb.grammar().parseString(token, parseAll=True)
    assert parsed[0].asDict() == expected


def test_hosts_selection():
    """A host selection is properly parsed and interpreted."""
    hosts = {'hosts': 'host[10-20,30-40].domain'}
    parsed = puppetdb.grammar().parseString(hosts['hosts'], parseAll=True)
    assert parsed[0].asDict() == hosts


class TestPuppetDBQueryV3:
    """PuppetDB backend query test class for API version 3."""

    def setup_method(self, _):
        """Set an instance of PuppetDBQuery for each test."""
        config = {'puppetdb': {'api_version': 3}}
        self.query = puppetdb.PuppetDBQuery(config)  # pylint: disable=attribute-defined-outside-init

    def test_instantiation(self):
        """An instance of PuppetDBQuery should be an instance of BaseQuery."""
        assert isinstance(self.query, BaseQuery)
        assert self.query.url == 'https://localhost:443/v3/'


class TestPuppetDBQueryV4:
    """PuppetDB backend query test class for API version 4."""

    def setup_method(self, _):
        """Set an instance of PuppetDBQuery for each test."""
        self.query = puppetdb.PuppetDBQuery({})  # pylint: disable=attribute-defined-outside-init

    def test_instantiation(self):
        """An instance of PuppetDBQuery should be an instance of BaseQuery."""
        assert isinstance(self.query, BaseQuery)
        assert self.query.url == 'https://localhost:443/pdb/query/v4/'

    def test_endpoint_getter(self):
        """Access to endpoint property should return nodes by default."""
        assert self.query.endpoint == 'nodes'

    @pytest.mark.parametrize('endpoint', set(puppetdb.PuppetDBQuery.endpoints.values()))
    def test_endpoint_setter_valid(self, endpoint):
        """Setting the endpoint property should accept valid values."""
        self.query.endpoint = endpoint
        assert self.query.endpoint == endpoint

    def test_endpoint_setter_invalid(self):
        """Setting the endpoint property should raise InvalidQueryError for an invalid value."""
        with pytest.raises(InvalidQueryError, match="Invalid value 'invalid_value'"):
            self.query.endpoint = 'invalid_value'

    def test_endpoint_setter_mixed1(self):
        """Setting the endpoint property twice to different values should raise InvalidQueryError (combination 1)."""
        assert self.query.endpoint == 'nodes'
        self.query.endpoint = 'resources'
        assert self.query.endpoint == 'resources'
        with pytest.raises(InvalidQueryError, match='Mixed endpoints are not supported'):
            self.query.endpoint = 'nodes'

    def test_endpoint_setter_mixed2(self):
        """Setting the endpoint property twice to different values should raise InvalidQueryError (combination 2)."""
        assert self.query.endpoint == 'nodes'
        self.query.endpoint = 'nodes'
        assert self.query.endpoint == 'nodes'
        with pytest.raises(InvalidQueryError, match='Mixed endpoints are not supported'):
            self.query.endpoint = 'resources'


def test_puppetdb_query_init_invalid():
    """Instantiating PuppetDBQuery with an unsupported API version should raise InvalidQueryError."""
    with pytest.raises(InvalidQueryError, match='Unsupported PuppetDB API version'):
        puppetdb.PuppetDBQuery({'puppetdb': {'api_version': 99}})


@mock.patch.object(puppetdb.PuppetDBQuery, '_api_call')
class TestPuppetDBQueryBuildV3:
    """PuppetDB backend API v3 query build test class."""

    def setup_method(self, _):
        """Set an instace of PuppetDBQuery for each test."""
        config = {'puppetdb': {'api_version': 3}}
        self.query = puppetdb.PuppetDBQuery(config)  # pylint: disable=attribute-defined-outside-init

    def test_add_category_resource_parameter_regex(self, mocked_api_call):
        """A resource's parameter query with a regex should raise InvalidQueryError."""
        with pytest.raises(InvalidQueryError, match='Regex operations are not supported in PuppetDB'):
            self.query.execute('R:resource%param ~ value.*')
            assert not mocked_api_call.called

    @pytest.mark.parametrize('query, expected', (
        (  # No hosts
            'host1!host1',
            ''),
        (  # Single host
            'host',
            '["or", ["=", "name", "host"]]'),
        (  # Multiple hosts
            'host[1-2]',
            '["or", ["=", "name", "host1"], ["=", "name", "host2"]]'),
        (  # Negated query
            'not host[1-2]',
            '["not", ["or", ["=", "name", "host1"], ["=", "name", "host2"]]]'),
        (  # Globbing hosts
            'host1*.domain',
            r'["or", ["~", "name", "^host1.*\\.domain$"]]'),
    ))
    def test_add_hosts(self, mocked_api_call, query, expected):
        """A host query should add the proper query token to the current_group."""
        self.query.execute(query)
        mocked_api_call.assert_called_with(expected)

    @pytest.mark.parametrize('query, expected', (
        (  # number
            'R:resource = 5551723',
            '["and", ["=", "type", "Resource"], ["=", "title", 5551723]]'),
        (  # number (neg)
            'R:resource = -23',
            '["and", ["=", "type", "Resource"], ["=", "title", -23]]'),
        (  # float
            'R:resource = 8675.309',
            '["and", ["=", "type", "Resource"], ["=", "title", 8675.309]]'),
        (  # hex
            'R:resource = 0x235A',
            '["and", ["=", "type", "Resource"], ["=", "title", 9050]]'),
        (  # lower hex
            'R:resource = 0x235a',
            '["and", ["=", "type", "Resource"], ["=", "title", 9050]]'),
        (  # oct
            'R:resource = 0777',
            '["and", ["=", "type", "Resource"], ["=", "title", 511]]'),
        (  # zero
            'R:resource = 0',
            '["and", ["=", "type", "Resource"], ["=", "title", 0]]'),
        (  # bareword
            'R:resource = true',
            '["and", ["=", "type", "Resource"], ["=", "title", true]]'),
        (  # quoted
            'R:resource = "value words"',
            '["and", ["=", "type", "Resource"], ["=", "title", "value words"]]'),
        (  # unquoted
            'R:resource = value',
            '["and", ["=", "type", "Resource"], ["=", "title", "value"]]'),
        (  # shortcut with numeric
            'O:role_name%param > 0x235A',
            ('["and", ["and", ["=", "type", "Class"], ["=", "title", "Role::Role_name"]], ["and", ["=", "type", "Class"'
             '], [">", ["parameter", "param"], 9050]]]')),
        (  # shortcut with bareword
            'O:role_name%param = true',
            ('["and", ["and", ["=", "type", "Class"], ["=", "title", "Role::Role_name"]], ["and", ["=", "type", "Class"'
             '], ["=", ["parameter", "param"], true]]]')),
    ))
    def test_types(self, mocked_api_call, query, expected):
        """A query should decode various types that puppetdb supports."""
        self.query.execute(query)
        mocked_api_call.assert_called_with(expected)

    @pytest.mark.parametrize('query, operator, expected', (
        (  # AND
            'host1 and host2',
            'and',
            '["and", ["or", ["=", "name", "host1"]], ["or", ["=", "name", "host2"]]]'),
        (  # OR
            'host1 or host2',
            'or',
            '["or", ["or", ["=", "name", "host1"]], ["or", ["=", "name", "host2"]]]'),
        (  # Multiple AND
            'host1 and host2 and host3',
            'and',
            '["and", ["or", ["=", "name", "host1"]], ["or", ["=", "name", "host2"]], ["or", ["=", "name", "host3"]]]'),
    ))
    def test_operator(self, mocked_api_call, query, operator, expected):
        """A query with boolean operators should set the boolean property to the current group."""
        self.query.execute(query)
        assert self.query.current_group['bool'] == operator
        mocked_api_call.assert_called_with(expected)


@mock.patch.object(puppetdb.PuppetDBQuery, '_api_call')
class TestPuppetDBQueryBuildV4:
    """PuppetDB backend API v4 query build test class."""

    def setup_method(self, _):
        """Set an instace of PuppetDBQuery for each test."""
        self.query = puppetdb.PuppetDBQuery({})  # pylint: disable=attribute-defined-outside-init

    @pytest.mark.parametrize('query, expected', (
        (  # Base fact
            'F:key=value',
            '["=", ["fact", "key"], "value"]'),
        (  # Negated
            'not F:key = value',
            '["not", ["=", ["fact", "key"], "value"]]'),
        (  # Different operator
            'F:key >= value',
            '[">=", ["fact", "key"], "value"]'),
        (  # Regex with backslash escaped
            r'F:key ~ value\\escaped',
            r'["~", ["fact", "key"], "value\\\\escaped"]'),
        (  # Regex with dot escaped
            r'F:key ~ value\.escaped',
            r'["~", ["fact", "key"], "value\\.escaped"]'),
    ))
    def test_add_category_fact(self, mocked_api_call, query, expected):
        """A fact query should add the proper query token to the current_group."""
        self.query.execute(query)
        mocked_api_call.assert_called_with(expected)

    @pytest.mark.parametrize('query, expected', (
        (  # Base resource equality
            'R:key = value',
            '["and", ["=", "type", "Key"], ["=", "title", "value"]]'),
        (  # Class title
            'R:class = classtitle',
            '["and", ["=", "type", "Class"], ["=", "title", "Classtitle"]]'),
        (  # Class path
            'R:class = resource::path::to::class',
            '["and", ["=", "type", "Class"], ["=", "title", "Resource::Path::To::Class"]]'),
        (  # Negated
            'not R:key = value',
            '["not", ["and", ["=", "type", "Key"], ["=", "title", "value"]]]'),
        (  # Regex backslash escaped
            r'R:key ~ value\\escaped',
            r'["and", ["=", "type", "Key"], ["~", "title", "value\\\\escaped"]]'),
        (  # Regex dot escaped
            r'R:key ~ value\.escaped',
            r'["and", ["=", "type", "Key"], ["~", "title", "value\\.escaped"]]'),
        (  # Regex class
            r'R:Class ~ "Role::(One|Another)"',
            r'["and", ["=", "type", "Class"], ["~", "title", "Role::(One|Another)"]]'),
        (  # Resource parameter
            'R:resource%param = value',
            '["and", ["=", "type", "Resource"], ["=", ["parameter", "param"], "value"]]'),
        (  # Resource parameter regex
            'R:resource%param ~ value.*',
            '["and", ["=", "type", "Resource"], ["~", ["parameter", "param"], "value.*"]]'),
        (  # Resource field
            'R:resource@field = value',
            '["and", ["=", "type", "Resource"], ["=", "field", "value"]]'),
        (  # Resource type
            'R:Resource',
            '["and", ["=", "type", "Resource"]]'),
        (  # Class shortcut
            'C:class_name',
            '["and", ["=", "type", "Class"], ["=", "title", "Class_name"]]'),
        (  # Class shortcut with path
            'C:module::class::name',
            '["and", ["=", "type", "Class"], ["=", "title", "Module::Class::Name"]]'),
        (  # Class shortcut with parameter
            'C:class_name%param = value',
            ('["and", ["and", ["=", "type", "Class"], ["=", "title", "Class_name"]], '
             '["and", ["=", "type", "Class"], ["=", ["parameter", "param"], "value"]]]')),
        (  # Class shortcut with field
            'C:class_name@field = value',
            ('["and", ["and", ["=", "type", "Class"], ["=", "title", "Class_name"]], '
             '["and", ["=", "type", "Class"], ["=", "field", "value"]]]')),
        (  # Profile shortcut
            'P:profile_name',
            '["and", ["=", "type", "Class"], ["=", "title", "Profile::Profile_name"]]'),
        (  # Profile shortcut path
            'P:module::name',
            '["and", ["=", "type", "Class"], ["=", "title", "Profile::Module::Name"]]'),
        (  # Profile shortcut with parameter
            'P:profile_name%param = value',
            ('["and", ["and", ["=", "type", "Class"], ["=", "title", "Profile::Profile_name"]], '
             '["and", ["=", "type", "Class"], ["=", ["parameter", "param"], "value"]]]')),
        (  # Profile shortcut with field
            'P:profile_name@field = value',
            ('["and", ["and", ["=", "type", "Class"], ["=", "title", "Profile::Profile_name"]], '
             '["and", ["=", "type", "Class"], ["=", "field", "value"]]]')),
        (  # Role shortcut
            'O:role_name',
            '["and", ["=", "type", "Class"], ["=", "title", "Role::Role_name"]]'),
        (  # Role shortcut path
            'O:module::name',
            '["and", ["=", "type", "Class"], ["=", "title", "Role::Module::Name"]]'),
        (  # Role shortcut with parameter
            'O:role_name%param = value',
            ('["and", ["and", ["=", "type", "Class"], ["=", "title", "Role::Role_name"]], '
             '["and", ["=", "type", "Class"], ["=", ["parameter", "param"], "value"]]]')),
        (  # Role shortcut with field
            'O:role_name@field = value',
            ('["and", ["and", ["=", "type", "Class"], ["=", "title", "Role::Role_name"]], '
             '["and", ["=", "type", "Class"], ["=", "field", "value"]]]')),
    ))
    def test_add_category_resource(self, mocked_api_call, query, expected):
        """A resource query should add the proper query token to the current_group."""
        self.query.execute(query)
        mocked_api_call.assert_called_with(expected)

    @pytest.mark.parametrize('query, message', (
        (  # Parameter and field
            'R:resource%param@field',
            'Resource key cannot contain both'),
        (  # Field and parameter
            'R:resource@field%param',
            'Resource key cannot contain both'),
        (  # Class shortcut with value
            'C:class_name = value',
            'The matching of a value is accepted only when using'),
        (  # Class shortcut with parameter and field
            'C:class_name%param@field',
            'Resource key cannot contain both'),
        (  # Class shortcut with field and parameter
            'C:class_name@field%param',
            'Resource key cannot contain both'),
        (  # Profile shortcut value
            'P:profile_name = value',
            'The matching of a value is accepted only when using'),
        (  # Profile shortcut with parameter and field
            'P:profile_name%param@field',
            'Resource key cannot contain both'),
        (  # Profile shortcut with field and parameter
            'P:profile_name@field%param',
            'Resource key cannot contain both'),
        (  # Role shortcut with value
            'O:role_name = value',
            'The matching of a value is accepted only when using'),
        (  # Role shortcut with parameter and field
            'O:role_name%param@field',
            'Resource key cannot contain both'),
        (  # Role shortcut with field and parameter
            'O:role_name@field%param',
            'Resource key cannot contain both'),
    ))
    def test_add_category_resource_raise(self, mocked_api_call, query, message):
        """A query with both a resource's parameter and field should raise InvalidQueryError."""
        with pytest.raises(InvalidQueryError, match=message):
            self.query.execute(query)
            assert not mocked_api_call.called

    @pytest.mark.parametrize('query, expected', (
        (  # No hosts
            'host1!host1',
            ''),
        (  # Single host
            'host',
            '["or", ["=", "certname", "host"]]'),
        (  # Multiple hosts
            'host[1-2]',
            '["or", ["=", "certname", "host1"], ["=", "certname", "host2"]]'),
        (  # Negated query
            'not host[1-2]',
            '["not", ["or", ["=", "certname", "host1"], ["=", "certname", "host2"]]]'),
        (  # Globbing hosts
            'host1*.domain',
            r'["or", ["~", "certname", "^host1.*\\.domain$"]]'),
    ))
    def test_add_hosts(self, mocked_api_call, query, expected):
        """A host query should add the proper query token to the current_group."""
        self.query.execute(query)
        mocked_api_call.assert_called_with(expected)

    @pytest.mark.parametrize('query, operator, expected', (
        (  # AND
            'host1 and host2',
            'and',
            '["and", ["or", ["=", "certname", "host1"]], ["or", ["=", "certname", "host2"]]]'),
        (  # OR
            'host1 or host2',
            'or',
            '["or", ["or", ["=", "certname", "host1"]], ["or", ["=", "certname", "host2"]]]'),
        (  # Multiple AND
            'host1 and host2 and host3',
            'and',
            (
                '["and", ["or", ["=", "certname", "host1"]], ["or", ["=", "certname", "host2"]], '
                '["or", ["=", "certname", "host3"]]]')),
    ))
    def test_operator(self, mocked_api_call, query, operator, expected):
        """A query with boolean operators should set the boolean property to the current group."""
        self.query.execute(query)
        assert self.query.current_group['bool'] == operator
        mocked_api_call.assert_called_with(expected)

    def test_and_or(self, mocked_api_call):
        """A query with 'and' and 'or' in the same group should raise InvalidQueryError."""
        with pytest.raises(InvalidQueryError, match='boolean operator, current operator was'):
            self.query.execute('host1 and host2 or host3')
            assert not mocked_api_call.called


@pytest.mark.parametrize('query, expected', (
    ('nodes_host[1-2]', 'nodes_host[1-2]'),  # Nodes
    ('R:Class = value', 'resources_host[1-2]'),  # Resources
    ('nodes_host1 or nodes_host2', 'nodes_host[1-2]'),  # Nodes with AND
    ('(nodes_host1 or nodes_host2)', 'nodes_host[1-2]'),  # Nodes with subgroup
    ('non_existent_host', None),  # No match
))
def test_endpoints(query_requests, query, expected):
    """Calling execute() with a query that goes to the nodes endpoint should return the list of hosts."""
    hosts = query_requests[0].execute(query)
    assert hosts == nodeset(expected)
    assert query_requests[1].call_count == 1


def test_error(query_requests):
    """Calling execute() if the request fails it should raise the requests exception."""
    with pytest.raises(HTTPError):
        query_requests[0].execute('invalid_query')
        assert query_requests[1].call_count == 1


def test_complex_query(query_requests):
    """Calling execute() with a complex query should return the exptected structure."""
    category = 'R'
    endpoint = query_requests[0].endpoints[category]
    key = query_requests[0].hosts_keys[endpoint]
    query_requests[1].register_uri('GET', query_requests[0].url + endpoint + '?query=', status_code=200, json=[
        {key: endpoint + '_host1', 'key': 'value1'}, {key: endpoint + '_host2', 'key': 'value2'}])

    hosts = query_requests[0].execute('(resources_host1 or resources_host2) and R:Class = MyClass')
    assert hosts == nodeset('resources_host[1-2]')
    assert query_requests[1].call_count == 1
